# Copyright (C) 2006, 2007, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests of the WorkingTree.unversion API."""

from breezy import errors, transport
from breezy.tests.matchers import HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestUnversion(TestCaseWithWorkingTree):

    def test_unversion_requires_write_lock(self):
        """WT.unversion([]) in a read lock raises ReadOnlyError."""
        tree = self.make_branch_and_tree('.')
        tree.lock_read()
        self.assertRaises(errors.ReadOnlyError, tree.unversion, [])
        tree.unlock()

    def test_unversion_missing_file(self):
        """WT.unversion(['missing']) raises NoSuchId."""
        tree = self.make_branch_and_tree('.')
        self.assertRaises(transport.NoSuchFile, tree.unversion, ['missing'])

    def test_unversion_parent_and_child_renamed_bug_187207(self):
        # When unversioning dirstate trees show a bug in dealing with
        # unversioning children of reparented children of unversioned
        # paths when relocation entries are present and the relocation
        # points later into the dirstate.
        tree = self.make_branch_and_tree('.')
        self.build_tree(['del/', 'del/sub/', 'del/sub/b'])
        tree.add(['del', 'del/sub', 'del/sub/b'])
        revid = tree.commit('setup')
        revtree = tree.branch.repository.revision_tree(revid)
        tree.rename_one('del/sub', 'sub')
        self.assertThat(
            tree,
            HasPathRelations(revtree, [('', ''), ('del/', 'del/'), ('sub/', 'del/sub/'), ('sub/b', 'del/sub/b')]))
        if tree.has_versioned_directories():
            tree.unversion(['del', 'sub/b'])
        else:
            tree.unversion(['sub/b'])
        self.assertThat(
            tree,
            HasPathRelations(revtree, [('', ''), ('sub/', 'del/sub/')]))

    def test_unversion_several_files(self):
        """After unversioning several files, they should not be versioned."""
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a', 'b', 'c'])
        tree.add(['a', 'b', 'c'])
        # within a lock unversion should take effect
        tree.lock_write()
        self.assertTrue(tree.is_versioned('a'))
        tree.unversion(['a', 'b'])
        self.assertFalse(tree.is_versioned('a'))
        self.assertFalse(tree.is_versioned('a'))
        self.assertFalse(tree.is_versioned('b'))
        self.assertTrue(tree.is_versioned('c'))
        self.assertTrue(tree.has_filename('a'))
        self.assertTrue(tree.has_filename('b'))
        self.assertTrue(tree.has_filename('c'))
        tree.unlock()
        # the changes should have persisted to disk - reopen the workingtree
        # to be sure.
        tree = tree.controldir.open_workingtree()
        self.addCleanup(tree.lock_read().unlock)
        self.assertFalse(tree.is_versioned('a'))
        self.assertFalse(tree.is_versioned('b'))
        self.assertTrue(tree.is_versioned('c'))
        self.assertTrue(tree.has_filename('a'))
        self.assertTrue(tree.has_filename('b'))
        self.assertTrue(tree.has_filename('c'))

    def test_unversion_subtree(self):
        """Unversioning the root of a subtree unversions the entire subtree."""
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a/', 'a/b', 'c'])
        tree.add(['a', 'a/b', 'c'])
        # within a lock unversion should take effect
        with tree.lock_write():
            tree.unversion(['a'])
            self.assertFalse(tree.is_versioned('a'))
            self.assertFalse(tree.is_versioned('a/b'))
            self.assertTrue(tree.is_versioned('c'))
            self.assertTrue(tree.has_filename('a'))
            self.assertTrue(tree.has_filename('a/b'))
            self.assertTrue(tree.has_filename('c'))

    def test_unversion_subtree_and_children(self):
        """Passing a child id will raise NoSuchId.

        This is because the parent directory will have already been removed.
        """
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a/', 'a/b', 'a/c', 'd'])
        tree.add(['a', 'a/b', 'a/c', 'd'])
        with tree.lock_write():
            tree.unversion(['a/b', 'a'])
            self.assertFalse(tree.is_versioned('a'))
            self.assertFalse(tree.is_versioned('a/b'))
            self.assertFalse(tree.is_versioned('a/c'))
            self.assertTrue(tree.is_versioned('d'))
            # The files are still on disk
            self.assertTrue(tree.has_filename('a'))
            self.assertTrue(tree.has_filename('a/b'))
            self.assertTrue(tree.has_filename('a/c'))
            self.assertTrue(tree.has_filename('d'))

    def test_unversion_renamed(self):
        tree = self.make_branch_and_tree('a')
        self.build_tree(['a/dir/', 'a/dir/f1', 'a/dir/f2', 'a/dir/f3',
                         'a/dir2/'])
        tree.add(['dir', 'dir/f1', 'dir/f2', 'dir/f3', 'dir2'])
        rev_id1 = tree.commit('init')
        revtree = tree.branch.repository.revision_tree(rev_id1)
        # Start off by renaming entries, and then unversion a bunch of entries
        # https://bugs.launchpad.net/bzr/+bug/114615
        tree.rename_one('dir/f1', 'dir/a')
        tree.rename_one('dir/f2', 'dir/z')
        tree.move(['dir/f3'], 'dir2')

        self.assertThat(
            tree, HasPathRelations(
                revtree,
                [('', ''), ('dir/', 'dir/'), ('dir2/', 'dir2/'),
                 ('dir/a', 'dir/f1'), ('dir/z', 'dir/f2'),
                 ('dir2/f3', 'dir/f3')]))

        tree.unversion({'dir'})
        self.assertThat(
            tree,
            HasPathRelations(
                revtree,
                [('', ''), ('dir2/', 'dir2/'), ('dir2/f3', 'dir/f3')]))

    def test_unversion_after_conflicted_merge(self):
        # Test for bug #114615
        tree_a = self.make_branch_and_tree('A')
        self.build_tree(['A/a/', 'A/a/m', 'A/a/n'])
        tree_a.add(['a', 'a/m', 'a/n'])
        tree_a.commit('init')

        tree_b = tree_a.controldir.sprout('B').open_workingtree()
        self.build_tree(['B/xyz/'])
        tree_b.add(['xyz'])
        tree_b.rename_one('a/m', 'xyz/m')
        tree_b.unversion(['a'])
        tree_b.commit('delete in B')

        self.assertThat(
            tree_b,
            HasPathRelations(
                tree_a,
                [('', ''), ('xyz/', None), ('xyz/m', 'a/m')]))

        self.build_tree_contents([('A/a/n', b'new contents for n\n')])
        tree_a.commit('change n in A')

        # Merging from A should introduce conflicts because 'n' was modified
        # and removed, so 'a' needs to be restored. We also have a conflict
        # because 'a' is still an existing directory
        conflicts = tree_b.merge_from_branch(tree_a.branch)
        if tree_b.has_versioned_directories():
            self.assertEqual(4, len(conflicts))
        else:
            self.assertEqual(1, len(conflicts))

        self.assertThat(
            tree_b,
            HasPathRelations(
                tree_a,
                [('', ''), ('a/', 'a/'), ('xyz/', None),
                 ('a/n.OTHER', 'a/n'), ('xyz/m', 'a/m')]))

        tree_b.unversion(['a'])
        self.assertThat(
            tree_b,
            HasPathRelations(
                tree_a,
                [('', ''), ('xyz/', None), ('xyz/m', 'a/m')]))



# Copyright (C) 2005-2011, 2013, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test the uncommit command."""

import os

from breezy import uncommit
from breezy.bzr.bzrdir import BzrDirMetaFormat1
from breezy.errors import BoundBranchOutOfDate
from breezy.tests import TestCaseWithTransport
from breezy.tests.script import ScriptRunner, run_script


class TestUncommit(TestCaseWithTransport):

    def create_simple_tree(self):
        wt = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a', 'tree/b', 'tree/c'])
        wt.add(['a', 'b', 'c'])
        wt.commit('initial commit', rev_id=b'a1')

        self.build_tree_contents([('tree/a', b'new contents of a\n')])
        wt.commit('second commit', rev_id=b'a2')

        return wt

    def test_uncommit(self):
        """Test uncommit functionality."""
        wt = self.create_simple_tree()

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --dry-run --force')
        self.assertContainsRe(out, 'Dry-run')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # Nothing has changed
        self.assertEqual([b'a2'], wt.get_parent_ids())

        # Uncommit, don't prompt
        out, err = self.run_bzr('uncommit --force')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # This should look like we are back in revno 1
        self.assertEqual([b'a1'], wt.get_parent_ids())
        out, err = self.run_bzr('status')
        self.assertEqual(out, 'modified:\n  a\n')

    def test_uncommit_interactive(self):
        """Uncommit seeks confirmation, and doesn't proceed without it."""
        wt = self.create_simple_tree()
        os.chdir('tree')
        run_script(self, """
        $ brz uncommit
        ...
        The above revision(s) will be removed.
        2>Uncommit these revisions? ([y]es, [n]o): no
        <n
        Canceled
        """)
        self.assertEqual([b'a2'], wt.get_parent_ids())

    def test_uncommit_no_history(self):
        wt = self.make_branch_and_tree('tree')
        out, err = self.run_bzr('uncommit --force', retcode=1)
        self.assertEqual('', err)
        self.assertEqual('No revisions to uncommit.\n', out)

    def test_uncommit_checkout(self):
        wt = self.create_simple_tree()
        checkout_tree = wt.branch.create_checkout('checkout')

        self.assertEqual([b'a2'], checkout_tree.get_parent_ids())

        os.chdir('checkout')
        out, err = self.run_bzr('uncommit --dry-run --force')
        self.assertContainsRe(out, 'Dry-run')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        self.assertEqual([b'a2'], checkout_tree.get_parent_ids())

        out, err = self.run_bzr('uncommit --force')
        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')

        # uncommit in a checkout should uncommit the parent branch
        # (but doesn't effect the other working tree)
        self.assertEqual([b'a1'], checkout_tree.get_parent_ids())
        self.assertEqual(b'a1', wt.branch.last_revision())
        self.assertEqual([b'a2'], wt.get_parent_ids())

    def test_uncommit_bound(self):
        os.mkdir('a')
        a = BzrDirMetaFormat1().initialize('a')
        a.create_repository()
        a.create_branch()
        t_a = a.create_workingtree()
        t_a.commit('commit 1')
        t_a.commit('commit 2')
        t_a.commit('commit 3')
        b = t_a.branch.create_checkout('b').branch
        uncommit.uncommit(b)
        self.assertEqual(b.last_revision_info()[0], 2)
        self.assertEqual(t_a.branch.last_revision_info()[0], 2)
        # update A's tree to not have the uncommitted revision referenced.
        t_a.update()
        t_a.commit('commit 3b')
        self.assertRaises(BoundBranchOutOfDate, uncommit.uncommit, b)
        b.pull(t_a.branch)
        uncommit.uncommit(b)

    def test_uncommit_bound_local(self):
        t_a = self.make_branch_and_tree('a')
        rev_id1 = t_a.commit('commit 1')
        rev_id2 = t_a.commit('commit 2')
        rev_id3 = t_a.commit('commit 3')
        b = t_a.branch.create_checkout('b').branch

        out, err = self.run_bzr(['uncommit', '--local', 'b', '--force'])
        self.assertEqual(rev_id3, t_a.last_revision())
        self.assertEqual((3, rev_id3), t_a.branch.last_revision_info())
        self.assertEqual((2, rev_id2), b.last_revision_info())

    def test_uncommit_revision(self):
        wt = self.create_simple_tree()

        os.chdir('tree')
        out, err = self.run_bzr('uncommit -r1 --force')

        self.assertNotContainsRe(out, 'initial commit')
        self.assertContainsRe(out, 'second commit')
        self.assertEqual([b'a1'], wt.get_parent_ids())
        self.assertEqual(b'a1', wt.branch.last_revision())

    def test_uncommit_neg_1(self):
        wt = self.create_simple_tree()
        os.chdir('tree')
        out, err = self.run_bzr('uncommit -r -1', retcode=1)
        self.assertEqual('No revisions to uncommit.\n', out)

    def test_uncommit_merges(self):
        wt = self.create_simple_tree()

        tree2 = wt.controldir.sprout('tree2').open_workingtree()

        tree2.commit('unchanged', rev_id=b'b3')
        tree2.commit('unchanged', rev_id=b'b4')

        wt.merge_from_branch(tree2.branch)
        wt.commit('merge b4', rev_id=b'a3')

        self.assertEqual([b'a3'], wt.get_parent_ids())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --force')

        self.assertEqual([b'a2', b'b4'], wt.get_parent_ids())

    def test_uncommit_pending_merge(self):
        wt = self.create_simple_tree()
        tree2 = wt.controldir.sprout('tree2').open_workingtree()
        tree2.commit('unchanged', rev_id=b'b3')

        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges([b'b3'])

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --force')
        self.assertEqual([b'a1', b'b3'], wt.get_parent_ids())

    def test_uncommit_multiple_merge(self):
        wt = self.create_simple_tree()

        tree2 = wt.controldir.sprout('tree2').open_workingtree()
        tree2.commit('unchanged', rev_id=b'b3')

        tree3 = wt.controldir.sprout('tree3').open_workingtree()
        tree3.commit('unchanged', rev_id=b'c3')

        wt.merge_from_branch(tree2.branch)
        wt.commit('merge b3', rev_id=b'a3')

        wt.merge_from_branch(tree3.branch)
        wt.commit('merge c3', rev_id=b'a4')

        self.assertEqual([b'a4'], wt.get_parent_ids())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --force -r 2')

        self.assertEqual([b'a2', b'b3', b'c3'], wt.get_parent_ids())

    def test_uncommit_merge_plus_pending(self):
        wt = self.create_simple_tree()

        tree2 = wt.controldir.sprout('tree2').open_workingtree()
        tree2.commit('unchanged', rev_id=b'b3')
        tree3 = wt.controldir.sprout('tree3').open_workingtree()
        tree3.commit('unchanged', rev_id=b'c3')

        wt.branch.fetch(tree2.branch)
        wt.set_pending_merges([b'b3'])
        wt.commit('merge b3', rev_id=b'a3')

        wt.merge_from_branch(tree3.branch)

        self.assertEqual([b'a3', b'c3'], wt.get_parent_ids())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --force -r 2')

        self.assertEqual([b'a2', b'b3', b'c3'], wt.get_parent_ids())

    def test_uncommit_shows_log_with_revision_id(self):
        wt = self.create_simple_tree()

        script = ScriptRunner()
        script.run_script(self, """
$ cd tree
$ brz uncommit --force
    2 ...
      second commit
...
The above revision(s) will be removed.
You can restore the old tip by running:
  brz pull . -r revid:a2
""")

    def test_uncommit_shows_pull_with_location(self):
        wt = self.create_simple_tree()

        script = ScriptRunner()
        script.run_script(self, """
$ brz uncommit --force tree
    2 ...
      second commit
...
The above revision(s) will be removed.
You can restore the old tip by running:
  brz pull -d tree tree -r revid:a2
""")

    def test_uncommit_octopus_merge(self):
        # Check that uncommit keeps the pending merges in the same order
        # though it will also filter out ones in the ancestry
        wt = self.create_simple_tree()

        tree2 = wt.controldir.sprout('tree2').open_workingtree()
        tree3 = wt.controldir.sprout('tree3').open_workingtree()

        tree2.commit('unchanged', rev_id=b'b3')
        tree3.commit('unchanged', rev_id=b'c3')

        wt.merge_from_branch(tree2.branch)
        wt.merge_from_branch(tree3.branch, force=True)
        wt.commit('merge b3, c3', rev_id=b'a3')

        tree2.commit('unchanged', rev_id=b'b4')
        tree3.commit('unchanged', rev_id=b'c4')

        wt.merge_from_branch(tree3.branch)
        wt.merge_from_branch(tree2.branch, force=True)
        wt.commit('merge b4, c4', rev_id=b'a4')

        self.assertEqual([b'a4'], wt.get_parent_ids())

        os.chdir('tree')
        out, err = self.run_bzr('uncommit --force -r 2')

        self.assertEqual([b'a2', b'c4', b'b4'], wt.get_parent_ids())

    def test_uncommit_nonascii(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('\u1234 message')
        out, err = self.run_bzr('uncommit --force tree', encoding='ascii')
        self.assertContainsRe(out, r'\? message')

    def test_uncommit_removes_tags(self):
        tree = self.make_branch_and_tree('tree')
        revid = tree.commit('message')
        tree.branch.tags.set_tag("atag", revid)
        out, err = self.run_bzr('uncommit --force tree')
        self.assertEqual({}, tree.branch.tags.get_tag_dict())

    def test_uncommit_keep_tags(self):
        tree = self.make_branch_and_tree('tree')
        revid = tree.commit('message')
        tree.branch.tags.set_tag("atag", revid)
        out, err = self.run_bzr('uncommit --keep-tags --force tree')
        self.assertEqual({"atag": revid}, tree.branch.tags.get_tag_dict())


class TestInconsistentDelta(TestCaseWithTransport):
    # See https://bugs.launchpad.net/bzr/+bug/855155
    # See https://bugs.launchpad.net/bzr/+bug/1100385
    # brz uncommit may result in error
    # 'An inconsistent delta was supplied involving'

    def test_inconsistent_delta(self):
        # Script taken from https://bugs.launchpad.net/bzr/+bug/855155/comments/26
        wt = self.make_branch_and_tree('test')
        self.build_tree(['test/a/', 'test/a/b', 'test/a/c'])
        wt.add(['a', 'a/b', 'a/c'])
        wt.commit('initial commit', rev_id=b'a1')
        wt.remove(['a/b', 'a/c'])
        wt.commit('remove b and c', rev_id=b'a2')
        self.run_bzr("uncommit --force test")



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz help.
"""


from breezy import config, i18n, tests
from breezy.tests.test_i18n import ZzzTranslations


class TestHelp(tests.TestCaseWithTransport):

    def test_help_basic(self):
        for cmd in ['--help', 'help', '-h', '-?']:
            output = self.run_bzr(cmd)[0]
            line1 = output.split('\n')[0]
            if not line1.startswith('Breezy'):
                self.fail("bad output from brz {}:\n{!r}".format(cmd, output))
        # see https://launchpad.net/products/bzr/+bug/35940, -h doesn't work

    def test_help_topics(self):
        """Smoketest for 'brz help topics'"""
        out, err = self.run_bzr('help topics')
        self.assertContainsRe(out, 'basic')
        self.assertContainsRe(out, 'topics')
        self.assertContainsRe(out, 'commands')
        self.assertContainsRe(out, 'revisionspec')

    def test_help_revisionspec(self):
        """Smoke test for 'brz help revisionspec'"""
        out, err = self.run_bzr('help revisionspec')
        self.assertContainsRe(out, 'revno:')
        self.assertContainsRe(out, 'date:')
        self.assertContainsRe(out, 'revid:')
        self.assertContainsRe(out, 'last:')
        self.assertContainsRe(out, 'before:')
        self.assertContainsRe(out, 'ancestor:')
        self.assertContainsRe(out, 'branch:')

    def test_help_checkouts(self):
        """Smoke test for 'brz help checkouts'"""
        out, err = self.run_bzr('help checkouts')
        self.assertContainsRe(out, 'checkout')
        self.assertContainsRe(out, 'lightweight')

    def test_help_urlspec(self):
        """Smoke test for 'brz help urlspec'"""
        out, err = self.run_bzr('help urlspec')
        self.assertContainsRe(out, 'bzr://')
        self.assertContainsRe(out, 'bzr\\+ssh://')
        self.assertContainsRe(out, 'file://')
        self.assertContainsRe(out, 'http://')
        self.assertContainsRe(out, 'https://')
        self.assertContainsRe(out, 'sftp://')

    def test_help_repositories(self):
        """Smoke test for 'brz help repositories'"""
        out, err = self.run_bzr('help repositories')
        from breezy.help_topics import _repositories, help_as_plain_text
        expected = help_as_plain_text(_repositories)
        self.assertEqual(expected, out)

    def test_help_working_trees(self):
        """Smoke test for 'brz help working-trees'"""
        out, err = self.run_bzr('help working-trees')
        from breezy.help_topics import _working_trees, help_as_plain_text
        expected = help_as_plain_text(_working_trees)
        self.assertEqual(expected, out)

    def test_help_status_flags(self):
        """Smoke test for 'brz help status-flags'"""
        out, err = self.run_bzr('help status-flags')
        from breezy.help_topics import _status_flags, help_as_plain_text
        expected = help_as_plain_text(_status_flags)
        self.assertEqual(expected, out)

    def test_help_commands(self):
        dash_help = self.run_bzr('--help commands')[0]
        commands = self.run_bzr('help commands')[0]
        hidden = self.run_bzr('help hidden-commands')[0]
        long_help = self.run_bzr('help --long')[0]
        qmark_long = self.run_bzr('? --long')[0]
        qmark_cmds = self.run_bzr('? commands')[0]
        self.assertEqual(dash_help, commands)
        self.assertEqual(dash_help, long_help)
        self.assertEqual(dash_help, qmark_long)
        self.assertEqual(dash_help, qmark_cmds)

    def test_help_width_zero(self):
        self.overrideEnv('BRZ_COLUMNS', '0')
        self.run_bzr('help commands')

    def test_hidden(self):
        help_commands = self.run_bzr('help commands')[0]
        help_hidden = self.run_bzr('help hidden-commands')[0]

        def extract_cmd_names(help_output):
            # keep only the command names to avoid matching on help text (there
            # is a high risk to fail a test when a plugin get installed
            # otherwise)
            cmds = []
            for line in help_output.split('\n'):
                if line.startswith(' '):
                    continue  # help on more than one line
                cmd = line.split(' ')[0]
                if line:
                    cmds.append(cmd)
            return cmds
        commands = extract_cmd_names(help_commands)
        hidden = extract_cmd_names(help_hidden)
        self.assertTrue('commit' in commands)
        self.assertTrue('commit' not in hidden)
        self.assertTrue('rocks' in hidden)
        self.assertTrue('rocks' not in commands)

    def test_help_detail(self):
        dash_h = self.run_bzr('diff -h')[0]
        help_x = self.run_bzr('help diff')[0]
        self.assertEqual(dash_h, help_x)
        self.assertContainsRe(help_x, "Purpose:")
        self.assertContainsRe(help_x, "Usage:")
        self.assertContainsRe(help_x, "Options:")
        self.assertContainsRe(help_x, "Description:")
        self.assertContainsRe(help_x, "Examples:")
        self.assertContainsRe(help_x, "See also:")
        self.assertContainsRe(help_x, "Aliases:")

    def test_help_usage(self):
        usage = self.run_bzr('diff --usage')[0]
        self.assertContainsRe(usage, "Purpose:")
        self.assertContainsRe(usage, "Usage:")
        self.assertContainsRe(usage, "Options:")
        self.assertNotContainsRe(usage, "Description:")
        self.assertNotContainsRe(usage, "Examples:")
        self.assertContainsRe(usage, "See also:")
        self.assertContainsRe(usage, "Aliases:")

    def test_help_help(self):
        help = self.run_bzr('help help')[0]
        qmark = self.run_bzr('? ?')[0]
        self.assertEqual(help, qmark)
        for line in help.split('\n'):
            if '--long' in line:
                self.assertContainsRe(line,
                                      r'Show help on all commands\.')

    def test_help_with_aliases(self):
        original = self.run_bzr('help cat')[0]

        conf = config.GlobalConfig.from_string('''[ALIASES]
c=cat
cat=cat
''', save=True)

        expected = original + "'brz cat' is an alias for 'brz cat'.\n"
        self.assertEqual(expected, self.run_bzr('help cat')[0])

        self.assertEqual("'brz c' is an alias for 'brz cat'.\n",
                         self.run_bzr('help c')[0])


class TestTranslatedHelp(tests.TestCaseWithTransport):
    """Tests for display of translated help topics"""

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, '_translations', ZzzTranslations())

    def test_help_command_utf8(self):
        out, err = self.run_bzr_raw(["help", "push"], encoding="utf-8")
        self.assertContainsRe(out, b"zz\xc3\xa5{{:See also:")

    def test_help_switch_utf8(self):
        out, err = self.run_bzr_raw(["push", "--help"], encoding="utf-8")
        self.assertContainsRe(out, b"zz\xc3\xa5{{:See also:")

    def test_help_command_ascii(self):
        out, err = self.run_bzr_raw(["help", "push"], encoding="ascii")
        self.assertContainsRe(out, b"zz\\?{{:See also:")

    def test_help_switch_ascii(self):
        out, err = self.run_bzr_raw(["push", "--help"], encoding="ascii")
        self.assertContainsRe(out, b"zz\\?{{:See also:")



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import time

from . import debug, errors, osutils, revision, trace

STEP_UNIQUE_SEARCHER_EVERY = 5

# DIAGRAM of terminology
#       A
#       /\
#      B  C
#      |  |\
#      D  E F
#      |\/| |
#      |/\|/
#      G  H
#
# In this diagram, relative to G and H:
# A, B, C, D, E are common ancestors.
# C, D and E are border ancestors, because each has a non-common descendant.
# D and E are least common ancestors because none of their descendants are
# common ancestors.
# C is not a least common ancestor because its descendant, E, is a common
# ancestor.
#
# The find_unique_lca algorithm will pick A in two steps:
# 1. find_lca('G', 'H') => ['D', 'E']
# 2. Since len(['D', 'E']) > 1, find_lca('D', 'E') => ['A']


class DictParentsProvider:
    """A parents provider for Graph objects."""

    def __init__(self, ancestry):
        self.ancestry = ancestry

    def __repr__(self):
        return 'DictParentsProvider(%r)' % self.ancestry

    # Note: DictParentsProvider does not implement get_cached_parent_map
    #       Arguably, the data is clearly cached in memory. However, this class
    #       is mostly used for testing, and it keeps the tests clean to not
    #       change it.

    def get_parent_map(self, keys):
        """See StackedParentsProvider.get_parent_map"""
        ancestry = self.ancestry
        return {k: ancestry[k] for k in keys if k in ancestry}


class StackedParentsProvider:
    """A parents provider which stacks (or unions) multiple providers.

    The providers are queries in the order of the provided parent_providers.
    """

    def __init__(self, parent_providers):
        self._parent_providers = parent_providers

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._parent_providers)

    def get_parent_map(self, keys):
        """Get a mapping of keys => parents

        A dictionary is returned with an entry for each key present in this
        source. If this source doesn't have information about a key, it should
        not include an entry.

        [NULL_REVISION] is used as the parent of the first user-committed
        revision.  Its parent list is empty.

        :param keys: An iterable returning keys to check (eg revision_ids)
        :return: A dictionary mapping each key to its parents
        """
        found = {}
        remaining = set(keys)
        # This adds getattr() overhead to each get_parent_map call. However,
        # this is StackedParentsProvider, which means we're dealing with I/O
        # (either local indexes, or remote RPCs), so CPU overhead should be
        # minimal.
        for parents_provider in self._parent_providers:
            get_cached = getattr(parents_provider, 'get_cached_parent_map',
                                 None)
            if get_cached is None:
                continue
            new_found = get_cached(remaining)
            found.update(new_found)
            remaining.difference_update(new_found)
            if not remaining:
                break
        if not remaining:
            return found
        for parents_provider in self._parent_providers:
            try:
                new_found = parents_provider.get_parent_map(remaining)
            except errors.UnsupportedOperation:
                continue
            found.update(new_found)
            remaining.difference_update(new_found)
            if not remaining:
                break
        return found


class CachingParentsProvider:
    """A parents provider which will cache the revision => parents as a dict.

    This is useful for providers which have an expensive look up.

    Either a ParentsProvider or a get_parent_map-like callback may be
    supplied.  If it provides extra un-asked-for parents, they will be cached,
    but filtered out of get_parent_map.

    The cache is enabled by default, but may be disabled and re-enabled.
    """

    def __init__(self, parent_provider=None, get_parent_map=None):
        """Constructor.

        :param parent_provider: The ParentProvider to use.  It or
            get_parent_map must be supplied.
        :param get_parent_map: The get_parent_map callback to use.  It or
            parent_provider must be supplied.
        """
        self._real_provider = parent_provider
        if get_parent_map is None:
            self._get_parent_map = self._real_provider.get_parent_map
        else:
            self._get_parent_map = get_parent_map
        self._cache = None
        self.enable_cache(True)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._real_provider)

    def enable_cache(self, cache_misses=True):
        """Enable cache."""
        if self._cache is not None:
            raise AssertionError('Cache enabled when already enabled.')
        self._cache = {}
        self._cache_misses = cache_misses
        self.missing_keys = set()

    def disable_cache(self):
        """Disable and clear the cache."""
        self._cache = None
        self._cache_misses = None
        self.missing_keys = set()

    def get_cached_map(self):
        """Return any cached get_parent_map values."""
        if self._cache is None:
            return None
        return dict(self._cache)

    def get_cached_parent_map(self, keys):
        """Return items from the cache.

        This returns the same info as get_parent_map, but explicitly does not
        invoke the supplied ParentsProvider to search for uncached values.
        """
        cache = self._cache
        if cache is None:
            return {}
        return {key: cache[key] for key in keys if key in cache}

    def get_parent_map(self, keys):
        """See StackedParentsProvider.get_parent_map."""
        cache = self._cache
        if cache is None:
            cache = self._get_parent_map(keys)
        else:
            needed_revisions = {key for key in keys if key not in cache}
            # Do not ask for negatively cached keys
            needed_revisions.difference_update(self.missing_keys)
            if needed_revisions:
                parent_map = self._get_parent_map(needed_revisions)
                cache.update(parent_map)
                if self._cache_misses:
                    for key in needed_revisions:
                        if key not in parent_map:
                            self.note_missing_key(key)
        result = {}
        for key in keys:
            value = cache.get(key)
            if value is not None:
                result[key] = value
        return result

    def note_missing_key(self, key):
        """Note that key is a missing key."""
        if self._cache_misses:
            self.missing_keys.add(key)


class CallableToParentsProviderAdapter:
    """A parents provider that adapts any callable to the parents provider API.

    i.e. it accepts calls to self.get_parent_map and relays them to the
    callable it was constructed with.
    """

    def __init__(self, a_callable):
        self.callable = a_callable

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.callable)

    def get_parent_map(self, keys):
        return self.callable(keys)


class Graph:
    """Provide incremental access to revision graphs.

    This is the generic implementation; it is intended to be subclassed to
    specialize it for other repository types.
    """

    def __init__(self, parents_provider):
        """Construct a Graph that uses several graphs as its input

        This should not normally be invoked directly, because there may be
        specialized implementations for particular repository types.  See
        Repository.get_graph().

        :param parents_provider: An object providing a get_parent_map call
            conforming to the behavior of
            StackedParentsProvider.get_parent_map.
        """
        if getattr(parents_provider, 'get_parents', None) is not None:
            self.get_parents = parents_provider.get_parents
        if getattr(parents_provider, 'get_parent_map', None) is not None:
            self.get_parent_map = parents_provider.get_parent_map
        self._parents_provider = parents_provider

    def __repr__(self):
        return 'Graph(%r)' % self._parents_provider

    def find_lca(self, *revisions):
        """Determine the lowest common ancestors of the provided revisions

        A lowest common ancestor is a common ancestor none of whose
        descendants are common ancestors.  In graphs, unlike trees, there may
        be multiple lowest common ancestors.

        This algorithm has two phases.  Phase 1 identifies border ancestors,
        and phase 2 filters border ancestors to determine lowest common
        ancestors.

        In phase 1, border ancestors are identified, using a breadth-first
        search starting at the bottom of the graph.  Searches are stopped
        whenever a node or one of its descendants is determined to be common

        In phase 2, the border ancestors are filtered to find the least
        common ancestors.  This is done by searching the ancestries of each
        border ancestor.

        Phase 2 is perfomed on the principle that a border ancestor that is
        not an ancestor of any other border ancestor is a least common
        ancestor.

        Searches are stopped when they find a node that is determined to be a
        common ancestor of all border ancestors, because this shows that it
        cannot be a descendant of any border ancestor.

        The scaling of this operation should be proportional to:

        1. The number of uncommon ancestors
        2. The number of border ancestors
        3. The length of the shortest path between a border ancestor and an
           ancestor of all border ancestors.
        """
        border_common, common, sides = self._find_border_ancestors(revisions)
        # We may have common ancestors that can be reached from each other.
        # - ask for the heads of them to filter it down to only ones that
        # cannot be reached from each other - phase 2.
        return self.heads(border_common)

    def find_difference(self, left_revision, right_revision):
        """Determine the graph difference between two revisions"""
        border, common, searchers = self._find_border_ancestors(
            [left_revision, right_revision])
        self._search_for_extra_common(common, searchers)
        left = searchers[0].seen
        right = searchers[1].seen
        return (left.difference(right), right.difference(left))

    def find_descendants(self, old_key, new_key):
        """Find descendants of old_key that are ancestors of new_key."""
        child_map = self.get_child_map(self._find_descendant_ancestors(
            old_key, new_key))
        graph = Graph(DictParentsProvider(child_map))
        searcher = graph._make_breadth_first_searcher([old_key])
        list(searcher)
        return searcher.seen

    def _find_descendant_ancestors(self, old_key, new_key):
        """Find ancestors of new_key that may be descendants of old_key."""
        stop = self._make_breadth_first_searcher([old_key])
        descendants = self._make_breadth_first_searcher([new_key])
        for revisions in descendants:
            old_stop = stop.seen.intersection(revisions)
            descendants.stop_searching_any(old_stop)
            seen_stop = descendants.find_seen_ancestors(stop.step())
            descendants.stop_searching_any(seen_stop)
        return descendants.seen.difference(stop.seen)

    def get_child_map(self, keys):
        """Get a mapping from parents to children of the specified keys.

        This is simply the inversion of get_parent_map.  Only supplied keys
        will be discovered as children.
        :return: a dict of key:child_list for keys.
        """
        parent_map = self._parents_provider.get_parent_map(keys)
        parent_child = {}
        for child, parents in sorted(parent_map.items()):
            for parent in parents:
                parent_child.setdefault(parent, []).append(child)
        return parent_child

    def find_distance_to_null(self, target_revision_id, known_revision_ids):
        """Find the left-hand distance to the NULL_REVISION.

        (This can also be considered the revno of a branch at
        target_revision_id.)

        :param target_revision_id: A revision_id which we would like to know
            the revno for.
        :param known_revision_ids: [(revision_id, revno)] A list of known
            revno, revision_id tuples. We'll use this to seed the search.
        """
        # Map from revision_ids to a known value for their revno
        known_revnos = dict(known_revision_ids)
        cur_tip = target_revision_id
        num_steps = 0
        NULL_REVISION = revision.NULL_REVISION
        known_revnos[NULL_REVISION] = 0

        searching_known_tips = list(known_revnos)

        unknown_searched = {}

        while cur_tip not in known_revnos:
            unknown_searched[cur_tip] = num_steps
            num_steps += 1
            to_search = {cur_tip}
            to_search.update(searching_known_tips)
            parent_map = self.get_parent_map(to_search)
            parents = parent_map.get(cur_tip, None)
            if not parents:  # An empty list or None is a ghost
                raise errors.GhostRevisionsHaveNoRevno(target_revision_id,
                                                       cur_tip)
            cur_tip = parents[0]
            next_known_tips = []
            for revision_id in searching_known_tips:
                parents = parent_map.get(revision_id, None)
                if not parents:
                    continue
                next = parents[0]
                next_revno = known_revnos[revision_id] - 1
                if next in unknown_searched:
                    # We have enough information to return a value right now
                    return next_revno + unknown_searched[next]
                if next in known_revnos:
                    continue
                known_revnos[next] = next_revno
                next_known_tips.append(next)
            searching_known_tips = next_known_tips

        # We reached a known revision, so just add in how many steps it took to
        # get there.
        return known_revnos[cur_tip] + num_steps

    def find_lefthand_distances(self, keys):
        """Find the distance to null for all the keys in keys.

        :param keys: keys to lookup.
        :return: A dict key->distance for all of keys.
        """
        # Optimisable by concurrent searching, but a random spread should get
        # some sort of hit rate.
        known_revnos = []
        ghosts = []
        for key in keys:
            try:
                known_revnos.append(
                    (key, self.find_distance_to_null(key, known_revnos)))
            except errors.GhostRevisionsHaveNoRevno:
                ghosts.append(key)
        for key in ghosts:
            known_revnos.append((key, -1))
        return dict(known_revnos)

    def find_unique_ancestors(self, unique_revision, common_revisions):
        """Find the unique ancestors for a revision versus others.

        This returns the ancestry of unique_revision, excluding all revisions
        in the ancestry of common_revisions. If unique_revision is in the
        ancestry, then the empty set will be returned.

        :param unique_revision: The revision_id whose ancestry we are
            interested in.
            (XXX: Would this API be better if we allowed multiple revisions on
            to be searched here?)
        :param common_revisions: Revision_ids of ancestries to exclude.
        :return: A set of revisions in the ancestry of unique_revision
        """
        if unique_revision in common_revisions:
            return set()

        # Algorithm description
        # 1) Walk backwards from the unique node and all common nodes.
        # 2) When a node is seen by both sides, stop searching it in the unique
        #    walker, include it in the common walker.
        # 3) Stop searching when there are no nodes left for the unique walker.
        #    At this point, you have a maximal set of unique nodes. Some of
        #    them may actually be common, and you haven't reached them yet.
        # 4) Start new searchers for the unique nodes, seeded with the
        #    information you have so far.
        # 5) Continue searching, stopping the common searches when the search
        #    tip is an ancestor of all unique nodes.
        # 6) Aggregate together unique searchers when they are searching the
        #    same tips. When all unique searchers are searching the same node,
        #    stop move it to a single 'all_unique_searcher'.
        # 7) The 'all_unique_searcher' represents the very 'tip' of searching.
        #    Most of the time this produces very little important information.
        #    So don't step it as quickly as the other searchers.
        # 8) Search is done when all common searchers have completed.

        unique_searcher, common_searcher = self._find_initial_unique_nodes(
            [unique_revision], common_revisions)

        unique_nodes = unique_searcher.seen.difference(common_searcher.seen)
        if not unique_nodes:
            return unique_nodes

        (all_unique_searcher,
         unique_tip_searchers) = self._make_unique_searchers(
             unique_nodes, unique_searcher, common_searcher)

        self._refine_unique_nodes(unique_searcher, all_unique_searcher,
                                  unique_tip_searchers, common_searcher)
        true_unique_nodes = unique_nodes.difference(common_searcher.seen)
        if 'graph' in debug.debug_flags:
            trace.mutter('Found %d truly unique nodes out of %d',
                         len(true_unique_nodes), len(unique_nodes))
        return true_unique_nodes

    def _find_initial_unique_nodes(self, unique_revisions, common_revisions):
        """Steps 1-3 of find_unique_ancestors.

        Find the maximal set of unique nodes. Some of these might actually
        still be common, but we are sure that there are no other unique nodes.

        :return: (unique_searcher, common_searcher)
        """

        unique_searcher = self._make_breadth_first_searcher(unique_revisions)
        # we know that unique_revisions aren't in common_revisions, so skip
        # past them.
        next(unique_searcher)
        common_searcher = self._make_breadth_first_searcher(common_revisions)

        # As long as we are still finding unique nodes, keep searching
        while unique_searcher._next_query:
            next_unique_nodes = set(unique_searcher.step())
            next_common_nodes = set(common_searcher.step())

            # Check if either searcher encounters new nodes seen by the other
            # side.
            unique_are_common_nodes = next_unique_nodes.intersection(
                common_searcher.seen)
            unique_are_common_nodes.update(
                next_common_nodes.intersection(unique_searcher.seen))
            if unique_are_common_nodes:
                ancestors = unique_searcher.find_seen_ancestors(
                    unique_are_common_nodes)
                # TODO: This is a bit overboard, we only really care about
                #       the ancestors of the tips because the rest we
                #       already know. This is *correct* but causes us to
                #       search too much ancestry.
                ancestors.update(
                    common_searcher.find_seen_ancestors(ancestors))
                unique_searcher.stop_searching_any(ancestors)
                common_searcher.start_searching(ancestors)

        return unique_searcher, common_searcher

    def _make_unique_searchers(self, unique_nodes, unique_searcher,
                               common_searcher):
        """Create a searcher for all the unique search tips (step 4).

        As a side effect, the common_searcher will stop searching any nodes
        that are ancestors of the unique searcher tips.

        :return: (all_unique_searcher, unique_tip_searchers)
        """
        unique_tips = self._remove_simple_descendants(
            unique_nodes, self.get_parent_map(unique_nodes))

        if len(unique_tips) == 1:
            unique_tip_searchers = []
            ancestor_all_unique = unique_searcher.find_seen_ancestors(
                unique_tips)
        else:
            unique_tip_searchers = []
            for tip in unique_tips:
                revs_to_search = unique_searcher.find_seen_ancestors([tip])
                revs_to_search.update(
                    common_searcher.find_seen_ancestors(revs_to_search))
                searcher = self._make_breadth_first_searcher(revs_to_search)
                # We don't care about the starting nodes.
                searcher._label = tip
                searcher.step()
                unique_tip_searchers.append(searcher)

            ancestor_all_unique = None
            for searcher in unique_tip_searchers:
                if ancestor_all_unique is None:
                    ancestor_all_unique = set(searcher.seen)
                else:
                    ancestor_all_unique = ancestor_all_unique.intersection(
                        searcher.seen)
        # Collapse all the common nodes into a single searcher
        all_unique_searcher = self._make_breadth_first_searcher(
            ancestor_all_unique)
        if ancestor_all_unique:
            # We've seen these nodes in all the searchers, so we'll just go to
            # the next
            all_unique_searcher.step()

            # Stop any search tips that are already known as ancestors of the
            # unique nodes
            stopped_common = common_searcher.stop_searching_any(
                common_searcher.find_seen_ancestors(ancestor_all_unique))

            total_stopped = 0
            for searcher in unique_tip_searchers:
                total_stopped += len(searcher.stop_searching_any(
                    searcher.find_seen_ancestors(ancestor_all_unique)))
        if 'graph' in debug.debug_flags:
            trace.mutter('For %d unique nodes, created %d + 1 unique searchers'
                         ' (%d stopped search tips, %d common ancestors'
                         ' (%d stopped common)',
                         len(unique_nodes), len(unique_tip_searchers),
                         total_stopped, len(ancestor_all_unique),
                         len(stopped_common))
        return all_unique_searcher, unique_tip_searchers

    def _step_unique_and_common_searchers(self, common_searcher,
                                          unique_tip_searchers,
                                          unique_searcher):
        """Step all the searchers"""
        newly_seen_common = set(common_searcher.step())
        newly_seen_unique = set()
        for searcher in unique_tip_searchers:
            next = set(searcher.step())
            next.update(unique_searcher.find_seen_ancestors(next))
            next.update(common_searcher.find_seen_ancestors(next))
            for alt_searcher in unique_tip_searchers:
                if alt_searcher is searcher:
                    continue
                next.update(alt_searcher.find_seen_ancestors(next))
            searcher.start_searching(next)
            newly_seen_unique.update(next)
        return newly_seen_common, newly_seen_unique

    def _find_nodes_common_to_all_unique(self, unique_tip_searchers,
                                         all_unique_searcher,
                                         newly_seen_unique, step_all_unique):
        """Find nodes that are common to all unique_tip_searchers.

        If it is time, step the all_unique_searcher, and add its nodes to the
        result.
        """
        common_to_all_unique_nodes = newly_seen_unique.copy()
        for searcher in unique_tip_searchers:
            common_to_all_unique_nodes.intersection_update(searcher.seen)
        common_to_all_unique_nodes.intersection_update(
            all_unique_searcher.seen)
        # Step all-unique less frequently than the other searchers.
        # In the common case, we don't need to spider out far here, so
        # avoid doing extra work.
        if step_all_unique:
            tstart = osutils.perf_counter()
            nodes = all_unique_searcher.step()
            common_to_all_unique_nodes.update(nodes)
            if 'graph' in debug.debug_flags:
                tdelta = osutils.perf_counter() - tstart
                trace.mutter('all_unique_searcher step() took %.3fs'
                             'for %d nodes (%d total), iteration: %s',
                             tdelta, len(nodes), len(all_unique_searcher.seen),
                             all_unique_searcher._iterations)
        return common_to_all_unique_nodes

    def _collapse_unique_searchers(self, unique_tip_searchers,
                                   common_to_all_unique_nodes):
        """Combine searchers that are searching the same tips.

        When two searchers are searching the same tips, we can stop one of the
        searchers. We also know that the maximal set of common ancestors is the
        intersection of the two original searchers.

        :return: A list of searchers that are searching unique nodes.
        """
        # Filter out searchers that don't actually search different
        # nodes. We already have the ancestry intersection for them
        unique_search_tips = {}
        for searcher in unique_tip_searchers:
            stopped = searcher.stop_searching_any(common_to_all_unique_nodes)
            will_search_set = frozenset(searcher._next_query)
            if not will_search_set:
                if 'graph' in debug.debug_flags:
                    trace.mutter('Unique searcher %s was stopped.'
                                 ' (%s iterations) %d nodes stopped',
                                 searcher._label,
                                 searcher._iterations,
                                 len(stopped))
            elif will_search_set not in unique_search_tips:
                # This searcher is searching a unique set of nodes, let it
                unique_search_tips[will_search_set] = [searcher]
            else:
                unique_search_tips[will_search_set].append(searcher)
        # TODO: it might be possible to collapse searchers faster when they
        #       only have *some* search tips in common.
        next_unique_searchers = []
        for searchers in unique_search_tips.values():
            if len(searchers) == 1:
                # Searching unique tips, go for it
                next_unique_searchers.append(searchers[0])
            else:
                # These searchers have started searching the same tips, we
                # don't need them to cover the same ground. The
                # intersection of their ancestry won't change, so create a
                # new searcher, combining their histories.
                next_searcher = searchers[0]
                for searcher in searchers[1:]:
                    next_searcher.seen.intersection_update(searcher.seen)
                if 'graph' in debug.debug_flags:
                    trace.mutter('Combining %d searchers into a single'
                                 ' searcher searching %d nodes with'
                                 ' %d ancestry',
                                 len(searchers),
                                 len(next_searcher._next_query),
                                 len(next_searcher.seen))
                next_unique_searchers.append(next_searcher)
        return next_unique_searchers

    def _refine_unique_nodes(self, unique_searcher, all_unique_searcher,
                             unique_tip_searchers, common_searcher):
        """Steps 5-8 of find_unique_ancestors.

        This function returns when common_searcher has stopped searching for
        more nodes.
        """
        # We step the ancestor_all_unique searcher only every
        # STEP_UNIQUE_SEARCHER_EVERY steps.
        step_all_unique_counter = 0
        # While we still have common nodes to search
        while common_searcher._next_query:
            (newly_seen_common,
             newly_seen_unique) = self._step_unique_and_common_searchers(
                common_searcher, unique_tip_searchers, unique_searcher)
            # These nodes are common ancestors of all unique nodes
            common_to_all_unique_nodes = self._find_nodes_common_to_all_unique(
                unique_tip_searchers, all_unique_searcher, newly_seen_unique,
                step_all_unique_counter == 0)
            step_all_unique_counter = ((step_all_unique_counter + 1)
                                       % STEP_UNIQUE_SEARCHER_EVERY)

            if newly_seen_common:
                # If a 'common' node is an ancestor of all unique searchers, we
                # can stop searching it.
                common_searcher.stop_searching_any(
                    all_unique_searcher.seen.intersection(newly_seen_common))
            if common_to_all_unique_nodes:
                common_to_all_unique_nodes.update(
                    common_searcher.find_seen_ancestors(
                        common_to_all_unique_nodes))
                # The all_unique searcher can start searching the common nodes
                # but everyone else can stop.
                # This is the sort of thing where we would like to not have it
                # start_searching all of the nodes, but only mark all of them
                # as seen, and have it search only the actual tips. Otherwise
                # it is another get_parent_map() traversal for it to figure out
                # what we already should know.
                all_unique_searcher.start_searching(common_to_all_unique_nodes)
                common_searcher.stop_searching_any(common_to_all_unique_nodes)

            next_unique_searchers = self._collapse_unique_searchers(
                unique_tip_searchers, common_to_all_unique_nodes)
            if len(unique_tip_searchers) != len(next_unique_searchers):
                if 'graph' in debug.debug_flags:
                    trace.mutter('Collapsed %d unique searchers => %d'
                                 ' at %s iterations',
                                 len(unique_tip_searchers),
                                 len(next_unique_searchers),
                                 all_unique_searcher._iterations)
            unique_tip_searchers = next_unique_searchers

    def get_parent_map(self, revisions):   # type: ignore
        """Get a map of key:parent_list for revisions.

        This implementation delegates to get_parents, for old parent_providers
        that do not supply get_parent_map.
        """
        result = {}
        for rev, parents in self.get_parents(revisions):
            if parents is not None:
                result[rev] = parents
        return result

    def _make_breadth_first_searcher(self, revisions):
        return _BreadthFirstSearcher(revisions, self)

    def _find_border_ancestors(self, revisions):
        """Find common ancestors with at least one uncommon descendant.

        Border ancestors are identified using a breadth-first
        search starting at the bottom of the graph.  Searches are stopped
        whenever a node or one of its descendants is determined to be common.

        This will scale with the number of uncommon ancestors.

        As well as the border ancestors, a set of seen common ancestors and a
        list of sets of seen ancestors for each input revision is returned.
        This allows calculation of graph difference from the results of this
        operation.
        """
        if None in revisions:
            raise errors.InvalidRevisionId(None, self)
        common_ancestors = set()
        searchers = [self._make_breadth_first_searcher([r])
                     for r in revisions]
        border_ancestors = set()

        while True:
            newly_seen = set()
            for searcher in searchers:
                new_ancestors = searcher.step()
                if new_ancestors:
                    newly_seen.update(new_ancestors)
            new_common = set()
            for revision in newly_seen:
                if revision in common_ancestors:
                    # Not a border ancestor because it was seen as common
                    # already
                    new_common.add(revision)
                    continue
                for searcher in searchers:
                    if revision not in searcher.seen:
                        break
                else:
                    # This is a border because it is a first common that we see
                    # after walking for a while.
                    border_ancestors.add(revision)
                    new_common.add(revision)
            if new_common:
                for searcher in searchers:
                    new_common.update(searcher.find_seen_ancestors(new_common))
                for searcher in searchers:
                    searcher.start_searching(new_common)
                common_ancestors.update(new_common)

            # Figure out what the searchers will be searching next, and if
            # there is only 1 set being searched, then we are done searching,
            # since all searchers would have to be searching the same data,
            # thus it *must* be in common.
            unique_search_sets = set()
            for searcher in searchers:
                will_search_set = frozenset(searcher._next_query)
                if will_search_set not in unique_search_sets:
                    # This searcher is searching a unique set of nodes, let it
                    unique_search_sets.add(will_search_set)

            if len(unique_search_sets) == 1:
                nodes = unique_search_sets.pop()
                uncommon_nodes = nodes.difference(common_ancestors)
                if uncommon_nodes:
                    raise AssertionError("Somehow we ended up converging"
                                         " without actually marking them as"
                                         " in common."
                                         "\nStart_nodes: %s"
                                         "\nuncommon_nodes: %s"
                                         % (revisions, uncommon_nodes))
                break
        return border_ancestors, common_ancestors, searchers

    def heads(self, keys):
        """Return the heads from amongst keys.

        This is done by searching the ancestries of each key.  Any key that is
        reachable from another key is not returned; all the others are.

        This operation scales with the relative depth between any two keys. If
        any two keys are completely disconnected all ancestry of both sides
        will be retrieved.

        :param keys: An iterable of keys.
        :return: A set of the heads. Note that as a set there is no ordering
            information. Callers will need to filter their input to create
            order if they need it.
        """
        candidate_heads = set(keys)
        if revision.NULL_REVISION in candidate_heads:
            # NULL_REVISION is only a head if it is the only entry
            candidate_heads.remove(revision.NULL_REVISION)
            if not candidate_heads:
                return {revision.NULL_REVISION}
        if len(candidate_heads) < 2:
            return candidate_heads
        searchers = {c: self._make_breadth_first_searcher([c])
                         for c in candidate_heads}
        active_searchers = dict(searchers)
        # skip over the actual candidate for each searcher
        for searcher in active_searchers.values():
            next(searcher)
        # The common walker finds nodes that are common to two or more of the
        # input keys, so that we don't access all history when a currently
        # uncommon search point actually meets up with something behind a
        # common search point. Common search points do not keep searches
        # active; they just allow us to make searches inactive without
        # accessing all history.
        common_walker = self._make_breadth_first_searcher([])
        while len(active_searchers) > 0:
            ancestors = set()
            # advance searches
            try:
                next(common_walker)
            except StopIteration:
                # No common points being searched at this time.
                pass
            for candidate in list(active_searchers):
                try:
                    searcher = active_searchers[candidate]
                except KeyError:
                    # rare case: we deleted candidate in a previous iteration
                    # through this for loop, because it was determined to be
                    # a descendant of another candidate.
                    continue
                try:
                    ancestors.update(next(searcher))
                except StopIteration:
                    del active_searchers[candidate]
                    continue
            # process found nodes
            new_common = set()
            for ancestor in ancestors:
                if ancestor in candidate_heads:
                    candidate_heads.remove(ancestor)
                    del searchers[ancestor]
                    if ancestor in active_searchers:
                        del active_searchers[ancestor]
                # it may meet up with a known common node
                if ancestor in common_walker.seen:
                    # some searcher has encountered our known common nodes:
                    # just stop it
                    ancestor_set = {ancestor}
                    for searcher in searchers.values():
                        searcher.stop_searching_any(ancestor_set)
                else:
                    # or it may have been just reached by all the searchers:
                    for searcher in searchers.values():
                        if ancestor not in searcher.seen:
                            break
                    else:
                        # The final active searcher has just reached this node,
                        # making it be known as a descendant of all candidates,
                        # so we can stop searching it, and any seen ancestors
                        new_common.add(ancestor)
                        for searcher in searchers.values():
                            seen_ancestors =\
                                searcher.find_seen_ancestors([ancestor])
                            searcher.stop_searching_any(seen_ancestors)
            common_walker.start_searching(new_common)
        return candidate_heads

    def find_merge_order(self, tip_revision_id, lca_revision_ids):
        """Find the order that each revision was merged into tip.

        This basically just walks backwards with a stack, and walks left-first
        until it finds a node to stop.
        """
        if len(lca_revision_ids) == 1:
            return list(lca_revision_ids)
        looking_for = set(lca_revision_ids)
        # TODO: Is there a way we could do this "faster" by batching up the
        # get_parent_map requests?
        # TODO: Should we also be culling the ancestry search right away? We
        # could add looking_for to the "stop" list, and walk their
        # ancestry in batched mode. The flip side is it might mean we walk a
        # lot of "stop" nodes, rather than only the minimum.
        # Then again, without it we may trace back into ancestry we could have
        # stopped early.
        stack = [tip_revision_id]
        found = []
        stop = set()
        while stack and looking_for:
            next = stack.pop()
            stop.add(next)
            if next in looking_for:
                found.append(next)
                looking_for.remove(next)
                if len(looking_for) == 1:
                    found.append(looking_for.pop())
                    break
                continue
            parent_ids = self.get_parent_map([next]).get(next, None)
            if not parent_ids:  # Ghost, nothing to search here
                continue
            for parent_id in reversed(parent_ids):
                # TODO: (performance) We see the parent at this point, but we
                #       wait to mark it until later to make sure we get left
                #       parents before right parents. However, instead of
                #       waiting until we have traversed enough parents, we
                #       could instead note that we've found it, and once all
                #       parents are in the stack, just reverse iterate the
                #       stack for them.
                if parent_id not in stop:
                    # this will need to be searched
                    stack.append(parent_id)
                stop.add(parent_id)
        return found

    def find_lefthand_merger(self, merged_key, tip_key):
        """Find the first lefthand ancestor of tip_key that merged merged_key.

        We do this by first finding the descendants of merged_key, then
        walking through the lefthand ancestry of tip_key until we find a key
        that doesn't descend from merged_key.  Its child is the key that
        merged merged_key.

        :return: The first lefthand ancestor of tip_key to merge merged_key.
            merged_key if it is a lefthand ancestor of tip_key.
            None if no ancestor of tip_key merged merged_key.
        """
        descendants = self.find_descendants(merged_key, tip_key)
        candidate_iterator = self.iter_lefthand_ancestry(tip_key)
        last_candidate = None
        for candidate in candidate_iterator:
            if candidate not in descendants:
                return last_candidate
            last_candidate = candidate

    def find_unique_lca(self, left_revision, right_revision,
                        count_steps=False):
        """Find a unique LCA.

        Find lowest common ancestors.  If there is no unique  common
        ancestor, find the lowest common ancestors of those ancestors.

        Iteration stops when a unique lowest common ancestor is found.
        The graph origin is necessarily a unique lowest common ancestor.

        Note that None is not an acceptable substitute for NULL_REVISION.
        in the input for this method.

        :param count_steps: If True, the return value will be a tuple of
            (unique_lca, steps) where steps is the number of times that
            find_lca was run.  If False, only unique_lca is returned.
        """
        revisions = [left_revision, right_revision]
        steps = 0
        while True:
            steps += 1
            lca = self.find_lca(*revisions)
            if len(lca) == 1:
                result = lca.pop()
                if count_steps:
                    return result, steps
                else:
                    return result
            if len(lca) == 0:
                raise errors.NoCommonAncestor(left_revision, right_revision)
            revisions = lca

    def iter_ancestry(self, revision_ids):
        """Iterate the ancestry of this revision.

        :param revision_ids: Nodes to start the search
        :return: Yield tuples mapping a revision_id to its parents for the
            ancestry of revision_id.
            Ghosts will be returned with None as their parents, and nodes
            with no parents will have NULL_REVISION as their only parent. (As
            defined by get_parent_map.)
            There will also be a node for (NULL_REVISION, ())
        """
        pending = set(revision_ids)
        processed = set()
        while pending:
            processed.update(pending)
            next_map = self.get_parent_map(pending)
            next_pending = set()
            for item in next_map.items():
                yield item
                next_pending.update(p for p in item[1] if p not in processed)
            ghosts = pending.difference(next_map)
            for ghost in ghosts:
                yield (ghost, None)
            pending = next_pending

    def iter_lefthand_ancestry(self, start_key, stop_keys=None):
        if stop_keys is None:
            stop_keys = ()
        next_key = start_key

        def get_parents(key):
            try:
                return self._parents_provider.get_parent_map([key])[key]
            except KeyError:
                raise errors.RevisionNotPresent(next_key, self)
        while True:
            if next_key in stop_keys:
                return
            parents = get_parents(next_key)
            yield next_key
            if len(parents) == 0:
                return
            else:
                next_key = parents[0]

    def iter_topo_order(self, revisions):
        """Iterate through the input revisions in topological order.

        This sorting only ensures that parents come before their children.
        An ancestor may sort after a descendant if the relationship is not
        visible in the supplied list of revisions.
        """
        from breezy import tsort
        sorter = tsort.TopoSorter(self.get_parent_map(revisions))
        return sorter.iter_topo_order()

    def is_ancestor(self, candidate_ancestor, candidate_descendant):
        """Determine whether a revision is an ancestor of another.

        We answer this using heads() as heads() has the logic to perform the
        smallest number of parent lookups to determine the ancestral
        relationship between N revisions.
        """
        return {candidate_descendant} == self.heads(
            [candidate_ancestor, candidate_descendant])

    def is_between(self, revid, lower_bound_revid, upper_bound_revid):
        """Determine whether a revision is between two others.

        returns true if and only if:
        lower_bound_revid <= revid <= upper_bound_revid
        """
        return ((upper_bound_revid is None or
                 self.is_ancestor(revid, upper_bound_revid)) and
                (lower_bound_revid is None or
                 self.is_ancestor(lower_bound_revid, revid)))

    def _search_for_extra_common(self, common, searchers):
        """Make sure that unique nodes are genuinely unique.

        After _find_border_ancestors, all nodes marked "common" are indeed
        common. Some of the nodes considered unique are not, due to history
        shortcuts stopping the searches early.

        We know that we have searched enough when all common search tips are
        descended from all unique (uncommon) nodes because we know that a node
        cannot be an ancestor of its own ancestor.

        :param common: A set of common nodes
        :param searchers: The searchers returned from _find_border_ancestors
        :return: None
        """
        # Basic algorithm...
        #   A) The passed in searchers should all be on the same tips, thus
        #      they should be considered the "common" searchers.
        #   B) We find the difference between the searchers, these are the
        #      "unique" nodes for each side.
        #   C) We do a quick culling so that we only start searching from the
        #      more interesting unique nodes. (A unique ancestor is more
        #      interesting than any of its children.)
        #   D) We start searching for ancestors common to all unique nodes.
        #   E) We have the common searchers stop searching any ancestors of
        #      nodes found by (D)
        #   F) When there are no more common search tips, we stop

        # TODO: We need a way to remove unique_searchers when they overlap with
        #       other unique searchers.
        if len(searchers) != 2:
            raise NotImplementedError(
                "Algorithm not yet implemented for > 2 searchers")
        common_searchers = searchers
        left_searcher = searchers[0]
        right_searcher = searchers[1]
        unique = left_searcher.seen.symmetric_difference(right_searcher.seen)
        if not unique:  # No unique nodes, nothing to do
            return
        total_unique = len(unique)
        unique = self._remove_simple_descendants(unique,
                                                 self.get_parent_map(unique))
        simple_unique = len(unique)

        unique_searchers = []
        for revision_id in unique:
            if revision_id in left_searcher.seen:
                parent_searcher = left_searcher
            else:
                parent_searcher = right_searcher
            revs_to_search = parent_searcher.find_seen_ancestors([revision_id])
            if not revs_to_search:  # XXX: This shouldn't be possible
                revs_to_search = [revision_id]
            searcher = self._make_breadth_first_searcher(revs_to_search)
            # We don't care about the starting nodes.
            searcher.step()
            unique_searchers.append(searcher)

        # possible todo: aggregate the common searchers into a single common
        #   searcher, just make sure that we include the nodes into the .seen
        #   properties of the original searchers

        ancestor_all_unique = None
        for searcher in unique_searchers:
            if ancestor_all_unique is None:
                ancestor_all_unique = set(searcher.seen)
            else:
                ancestor_all_unique = ancestor_all_unique.intersection(
                    searcher.seen)

        trace.mutter('Started %d unique searchers for %d unique revisions',
                     simple_unique, total_unique)

        while True:  # If we have no more nodes we have nothing to do
            newly_seen_common = set()
            for searcher in common_searchers:
                newly_seen_common.update(searcher.step())
            newly_seen_unique = set()
            for searcher in unique_searchers:
                newly_seen_unique.update(searcher.step())
            new_common_unique = set()
            for revision in newly_seen_unique:
                for searcher in unique_searchers:
                    if revision not in searcher.seen:
                        break
                else:
                    # This is a border because it is a first common that we see
                    # after walking for a while.
                    new_common_unique.add(revision)
            if newly_seen_common:
                # These are nodes descended from one of the 'common' searchers.
                # Make sure all searchers are on the same page
                for searcher in common_searchers:
                    newly_seen_common.update(
                        searcher.find_seen_ancestors(newly_seen_common))
                # We start searching the whole ancestry. It is a bit wasteful,
                # though. We really just want to mark all of these nodes as
                # 'seen' and then start just the tips. However, it requires a
                # get_parent_map() call to figure out the tips anyway, and all
                # redundant requests should be fairly fast.
                for searcher in common_searchers:
                    searcher.start_searching(newly_seen_common)

                # If a 'common' node is an ancestor of all unique searchers, we
                # can stop searching it.
                stop_searching_common = ancestor_all_unique.intersection(
                    newly_seen_common)
                if stop_searching_common:
                    for searcher in common_searchers:
                        searcher.stop_searching_any(stop_searching_common)
            if new_common_unique:
                # We found some ancestors that are common
                for searcher in unique_searchers:
                    new_common_unique.update(
                        searcher.find_seen_ancestors(new_common_unique))
                # Since these are common, we can grab another set of ancestors
                # that we have seen
                for searcher in common_searchers:
                    new_common_unique.update(
                        searcher.find_seen_ancestors(new_common_unique))

                # We can tell all of the unique searchers to start at these
                # nodes, and tell all of the common searchers to *stop*
                # searching these nodes
                for searcher in unique_searchers:
                    searcher.start_searching(new_common_unique)
                for searcher in common_searchers:
                    searcher.stop_searching_any(new_common_unique)
                ancestor_all_unique.update(new_common_unique)

                # Filter out searchers that don't actually search different
                # nodes. We already have the ancestry intersection for them
                next_unique_searchers = []
                unique_search_sets = set()
                for searcher in unique_searchers:
                    will_search_set = frozenset(searcher._next_query)
                    if will_search_set not in unique_search_sets:
                        # This searcher is searching a unique set of nodes, let
                        # it
                        unique_search_sets.add(will_search_set)
                        next_unique_searchers.append(searcher)
                unique_searchers = next_unique_searchers
            for searcher in common_searchers:
                if searcher._next_query:
                    break
            else:
                # All common searcher have stopped searching
                return

    def _remove_simple_descendants(self, revisions, parent_map):
        """remove revisions which are children of other ones in the set

        This doesn't do any graph searching, it just checks the immediate
        parent_map to find if there are any children which can be removed.

        :param revisions: A set of revision_ids
        :return: A set of revision_ids with the children removed
        """
        simple_ancestors = revisions.copy()
        # TODO: jam 20071214 we *could* restrict it to searching only the
        #       parent_map of revisions already present in 'revisions', but
        #       considering the general use case, I think this is actually
        #       better.

        # This is the same as the following loop. I don't know that it is any
        # faster.
        # simple_ancestors.difference_update(r for r, p_ids in parent_map.iteritems()
        # if p_ids is not None and revisions.intersection(p_ids))
        # return simple_ancestors

        # Yet Another Way, invert the parent map (which can be cached)
        ## descendants = {}
        # for revision_id, parent_ids in parent_map.iteritems():
        # for p_id in parent_ids:
        ##       descendants.setdefault(p_id, []).append(revision_id)
        # for revision in revisions.intersection(descendants):
        # simple_ancestors.difference_update(descendants[revision])
        # return simple_ancestors
        for revision, parent_ids in parent_map.items():
            if parent_ids is None:
                continue
            for parent_id in parent_ids:
                if parent_id in revisions:
                    # This node has a parent present in the set, so we can
                    # remove it
                    simple_ancestors.discard(revision)
                    break
        return simple_ancestors


class HeadsCache:
    """A cache of results for graph heads calls."""

    def __init__(self, graph):
        self.graph = graph
        self._heads = {}

    def heads(self, keys):
        """Return the heads of keys.

        This matches the API of Graph.heads(), specifically the return value is
        a set which can be mutated, and ordering of the input is not preserved
        in the output.

        :see also: Graph.heads.
        :param keys: The keys to calculate heads for.
        :return: A set containing the heads, which may be mutated without
            affecting future lookups.
        """
        keys = frozenset(keys)
        try:
            return set(self._heads[keys])
        except KeyError:
            heads = self.graph.heads(keys)
            self._heads[keys] = heads
            return set(heads)


class FrozenHeadsCache:
    """Cache heads() calls, assuming the caller won't modify them."""

    def __init__(self, graph):
        self.graph = graph
        self._heads = {}

    def heads(self, keys):
        """Return the heads of keys.

        Similar to Graph.heads(). The main difference is that the return value
        is a frozen set which cannot be mutated.

        :see also: Graph.heads.
        :param keys: The keys to calculate heads for.
        :return: A frozenset containing the heads.
        """
        keys = frozenset(keys)
        try:
            return self._heads[keys]
        except KeyError:
            heads = frozenset(self.graph.heads(keys))
            self._heads[keys] = heads
            return heads

    def cache(self, keys, heads):
        """Store a known value."""
        self._heads[frozenset(keys)] = frozenset(heads)


class _BreadthFirstSearcher:
    """Parallel search breadth-first the ancestry of revisions.

    This class implements the iterator protocol, but additionally
    1. provides a set of seen ancestors, and
    2. allows some ancestries to be unsearched, via stop_searching_any
    """

    def __init__(self, revisions, parents_provider):
        self._iterations = 0
        self._next_query = set(revisions)
        self.seen = set()
        self._started_keys = set(self._next_query)
        self._stopped_keys = set()
        self._parents_provider = parents_provider
        self._returning = 'next_with_ghosts'
        self._current_present = set()
        self._current_ghosts = set()
        self._current_parents = {}

    def __repr__(self):
        if self._iterations:
            prefix = "searching"
        else:
            prefix = "starting"
        search = '{}={!r}'.format(prefix, list(self._next_query))
        return ('_BreadthFirstSearcher(iterations=%d, %s,'
                ' seen=%r)' % (self._iterations, search, list(self.seen)))

    def get_state(self):
        """Get the current state of this searcher.

        :return: Tuple with started keys, excludes and included keys
        """
        if self._returning == 'next':
            # We have to know the current nodes children to be able to list the
            # exclude keys for them. However, while we could have a second
            # look-ahead result buffer and shuffle things around, this method
            # is typically only called once per search - when memoising the
            # results of the search.
            found, ghosts, next, parents = self._do_query(self._next_query)
            # pretend we didn't query: perhaps we should tweak _do_query to be
            # entirely stateless?
            self.seen.difference_update(next)
            next_query = next.union(ghosts)
        else:
            next_query = self._next_query
        excludes = self._stopped_keys.union(next_query)
        included_keys = self.seen.difference(excludes)
        return self._started_keys, excludes, included_keys

    def step(self):
        try:
            return next(self)
        except StopIteration:
            return ()

    def __next__(self):
        """Return the next ancestors of this revision.

        Ancestors are returned in the order they are seen in a breadth-first
        traversal.  No ancestor will be returned more than once. Ancestors are
        returned before their parentage is queried, so ghosts and missing
        revisions (including the start revisions) are included in the result.
        This can save a round trip in LCA style calculation by allowing
        convergence to be detected without reading the data for the revision
        the convergence occurs on.

        :return: A set of revision_ids.
        """
        if self._returning != 'next':
            # switch to returning the query, not the results.
            self._returning = 'next'
            self._iterations += 1
        else:
            self._advance()
        if len(self._next_query) == 0:
            raise StopIteration()
        # We have seen what we're querying at this point as we are returning
        # the query, not the results.
        self.seen.update(self._next_query)
        return self._next_query

    next = __next__

    def next_with_ghosts(self):
        """Return the next found ancestors, with ghosts split out.

        Ancestors are returned in the order they are seen in a breadth-first
        traversal.  No ancestor will be returned more than once. Ancestors are
        returned only after asking for their parents, which allows us to detect
        which revisions are ghosts and which are not.

        :return: A tuple with (present ancestors, ghost ancestors) sets.
        """
        if self._returning != 'next_with_ghosts':
            # switch to returning the results, not the current query.
            self._returning = 'next_with_ghosts'
            self._advance()
        if len(self._next_query) == 0:
            raise StopIteration()
        self._advance()
        return self._current_present, self._current_ghosts

    def _advance(self):
        """Advance the search.

        Updates self.seen, self._next_query, self._current_present,
        self._current_ghosts, self._current_parents and self._iterations.
        """
        self._iterations += 1
        found, ghosts, next, parents = self._do_query(self._next_query)
        self._current_present = found
        self._current_ghosts = ghosts
        self._next_query = next
        self._current_parents = parents
        # ghosts are implicit stop points, otherwise the search cannot be
        # repeated when ghosts are filled.
        self._stopped_keys.update(ghosts)

    def _do_query(self, revisions):
        """Query for revisions.

        Adds revisions to the seen set.

        :param revisions: Revisions to query.
        :return: A tuple: (set(found_revisions), set(ghost_revisions),
           set(parents_of_found_revisions), dict(found_revisions:parents)).
        """
        found_revisions = set()
        parents_of_found = set()
        # revisions may contain nodes that point to other nodes in revisions:
        # we want to filter them out.
        seen = self.seen
        seen.update(revisions)
        parent_map = self._parents_provider.get_parent_map(revisions)
        found_revisions.update(parent_map)
        for rev_id, parents in parent_map.items():
            if parents is None:
                continue
            new_found_parents = [p for p in parents if p not in seen]
            if new_found_parents:
                # Calling set.update() with an empty generator is actually
                # rather expensive.
                parents_of_found.update(new_found_parents)
        ghost_revisions = revisions - found_revisions
        return found_revisions, ghost_revisions, parents_of_found, parent_map

    def __iter__(self):
        return self

    def find_seen_ancestors(self, revisions):
        """Find ancestors of these revisions that have already been seen.

        This function generally makes the assumption that querying for the
        parents of a node that has already been queried is reasonably cheap.
        (eg, not a round trip to a remote host).
        """
        # TODO: Often we might ask one searcher for its seen ancestors, and
        #       then ask another searcher the same question. This can result in
        #       searching the same revisions repeatedly if the two searchers
        #       have a lot of overlap.
        all_seen = self.seen
        pending = set(revisions).intersection(all_seen)
        seen_ancestors = set(pending)

        if self._returning == 'next':
            # self.seen contains what nodes have been returned, not what nodes
            # have been queried. We don't want to probe for nodes that haven't
            # been searched yet.
            not_searched_yet = self._next_query
        else:
            not_searched_yet = ()
        pending.difference_update(not_searched_yet)
        get_parent_map = self._parents_provider.get_parent_map
        while pending:
            parent_map = get_parent_map(pending)
            all_parents = []
            # We don't care if it is a ghost, since it can't be seen if it is
            # a ghost
            for parent_ids in parent_map.values():
                all_parents.extend(parent_ids)
            next_pending = all_seen.intersection(
                all_parents).difference(seen_ancestors)
            seen_ancestors.update(next_pending)
            next_pending.difference_update(not_searched_yet)
            pending = next_pending

        return seen_ancestors

    def stop_searching_any(self, revisions):
        """
        Remove any of the specified revisions from the search list.

        None of the specified revisions are required to be present in the
        search list.

        It is okay to call stop_searching_any() for revisions which were seen
        in previous iterations. It is the callers responsibility to call
        find_seen_ancestors() to make sure that current search tips that are
        ancestors of those revisions are also stopped.  All explicitly stopped
        revisions will be excluded from the search result's get_keys(), though.
        """
        # TODO: does this help performance?
        # if not revisions:
        #     return set()
        revisions = frozenset(revisions)
        if self._returning == 'next':
            stopped = self._next_query.intersection(revisions)
            self._next_query = self._next_query.difference(revisions)
        else:
            stopped_present = self._current_present.intersection(revisions)
            stopped = stopped_present.union(
                self._current_ghosts.intersection(revisions))
            self._current_present.difference_update(stopped)
            self._current_ghosts.difference_update(stopped)
            # stopping 'x' should stop returning parents of 'x', but
            # not if 'y' always references those same parents
            stop_rev_references = {}
            for rev in stopped_present:
                for parent_id in self._current_parents[rev]:
                    if parent_id not in stop_rev_references:
                        stop_rev_references[parent_id] = 0
                    stop_rev_references[parent_id] += 1
            # if only the stopped revisions reference it, the ref count will be
            # 0 after this loop
            for parents in self._current_parents.values():
                for parent_id in parents:
                    try:
                        stop_rev_references[parent_id] -= 1
                    except KeyError:
                        pass
            stop_parents = set()
            for rev_id, refs in stop_rev_references.items():
                if refs == 0:
                    stop_parents.add(rev_id)
            self._next_query.difference_update(stop_parents)
        self._stopped_keys.update(stopped)
        self._stopped_keys.update(revisions)
        return stopped

    def start_searching(self, revisions):
        """Add revisions to the search.

        The parents of revisions will be returned from the next call to next()
        or next_with_ghosts(). If next_with_ghosts was the most recently used
        next* call then the return value is the result of looking up the
        ghost/not ghost status of revisions. (A tuple (present, ghosted)).
        """
        revisions = frozenset(revisions)
        self._started_keys.update(revisions)
        new_revisions = revisions.difference(self.seen)
        if self._returning == 'next':
            self._next_query.update(new_revisions)
            self.seen.update(new_revisions)
        else:
            # perform a query on revisions
            revs, ghosts, query, parents = self._do_query(revisions)
            self._stopped_keys.update(ghosts)
            self._current_present.update(revs)
            self._current_ghosts.update(ghosts)
            self._next_query.update(query)
            self._current_parents.update(parents)
            return revs, ghosts


def invert_parent_map(parent_map):
    """Given a map from child => parents, create a map of parent=>children"""
    child_map = {}
    for child, parents in parent_map.items():
        for p in parents:
            # Any given parent is likely to have only a small handful
            # of children, many will have only one. So we avoid mem overhead of
            # a list, in exchange for extra copying of tuples
            if p not in child_map:
                child_map[p] = (child,)
            else:
                child_map[p] = child_map[p] + (child,)
    return child_map


def collapse_linear_regions(parent_map):
    """Collapse regions of the graph that are 'linear'.

    For example::

      A:[B], B:[C]

    can be collapsed by removing B and getting::

      A:[C]

    :param parent_map: A dictionary mapping children to their parents
    :return: Another dictionary with 'linear' chains collapsed
    """
    # Note: this isn't a strictly minimal collapse. For example:
    #   A
    #  / \
    # B   C
    #  \ /
    #   D
    #   |
    #   E
    # Will not have 'D' removed, even though 'E' could fit. Also:
    #   A
    #   |    A
    #   B => |
    #   |    C
    #   C
    # A and C are both kept because they are edges of the graph. We *could* get
    # rid of A if we wanted.
    #   A
    #  / \
    # B   C
    # |   |
    # D   E
    #  \ /
    #   F
    # Will not have any nodes removed, even though you do have an
    # 'uninteresting' linear D->B and E->C
    children = {}
    for child, parents in parent_map.items():
        children.setdefault(child, [])
        for p in parents:
            children.setdefault(p, []).append(child)

    removed = set()
    result = dict(parent_map)
    for node in parent_map:
        parents = result[node]
        if len(parents) == 1:
            parent_children = children[parents[0]]
            if len(parent_children) != 1:
                # This is not the only child
                continue
            node_children = children[node]
            if len(node_children) != 1:
                continue
            child_parents = result.get(node_children[0], None)
            if len(child_parents) != 1:
                # This is not its only parent
                continue
            # The child of this node only points at it, and the parent only has
            # this as a child. remove this node, and join the others together
            result[node_children[0]] = parents
            children[parents[0]] = node_children
            del result[node]
            del children[node]
            removed.add(node)

    return result


class GraphThunkIdsToKeys:
    """Forwards calls about 'ids' to be about keys internally."""

    def __init__(self, graph):
        self._graph = graph

    def topo_sort(self):
        return [r for (r,) in self._graph.topo_sort()]

    def heads(self, ids):
        """See Graph.heads()"""
        as_keys = [(i,) for i in ids]
        head_keys = self._graph.heads(as_keys)
        return {h[0] for h in head_keys}

    def merge_sort(self, tip_revision):
        nodes = self._graph.merge_sort((tip_revision,))
        for node in nodes:
            node.key = node.key[0]
        return nodes

    def add_node(self, revision, parents):
        self._graph.add_node((revision,), [(p,) for p in parents])


_counters = [0, 0, 0, 0, 0, 0, 0]
try:
    from ._known_graph_pyx import KnownGraph
except ImportError as e:
    osutils.failed_to_load_extension(e)
    from ._known_graph_py import KnownGraph  # noqa: F401



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""HTTPS test server, available when ssl python module is available"""

import ssl

from . import http_server, ssl_certs, test_server


class TestingHTTPSServerMixin:

    def __init__(self, key_file, cert_file):
        self.key_file = key_file
        self.cert_file = cert_file

    def _get_ssl_request(self, sock, addr):
        """Wrap the socket with SSL"""
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        if self.cert_file:
            ssl_context.load_cert_chain(self.cert_file, self.key_file)
        ssl_sock = ssl_context.wrap_socket(
            sock=sock, server_side=True,
            do_handshake_on_connect=False)
        return ssl_sock, addr

    def verify_request(self, request, client_address):
        """Verify the request.

        Return True if we should proceed with this request, False if we should
        not even touch a single byte in the socket !
        """
        serving = test_server.TestingTCPServerMixin.verify_request(
            self, request, client_address)
        if serving:
            try:
                request.do_handshake()
            except ssl.SSLError:
                # FIXME: We proabaly want more tests to capture which ssl
                # errors are worth reporting but mostly our tests want an https
                # server that works -- vila 2012-01-19
                return False
        return serving

    def ignored_exceptions_during_shutdown(self, e):
        base = test_server.TestingTCPServerMixin
        return base.ignored_exceptions_during_shutdown(self, e)


class TestingHTTPSServer(TestingHTTPSServerMixin,
                         http_server.TestingHTTPServer):

    def __init__(self, server_address, request_handler_class,
                 test_case_server, key_file, cert_file):
        TestingHTTPSServerMixin.__init__(self, key_file, cert_file)
        http_server.TestingHTTPServer.__init__(
            self, server_address, request_handler_class, test_case_server)

    def get_request(self):
        sock, addr = http_server.TestingHTTPServer.get_request(self)
        return self._get_ssl_request(sock, addr)


class TestingThreadingHTTPSServer(TestingHTTPSServerMixin,
                                  http_server.TestingThreadingHTTPServer):

    def __init__(self, server_address, request_handler_class,
                 test_case_server, key_file, cert_file):
        TestingHTTPSServerMixin.__init__(self, key_file, cert_file)
        http_server.TestingThreadingHTTPServer.__init__(
            self, server_address, request_handler_class, test_case_server)

    def get_request(self):
        sock, addr = http_server.TestingThreadingHTTPServer.get_request(self)
        return self._get_ssl_request(sock, addr)


class HTTPSServer(http_server.HttpServer):

    _url_protocol = 'https'

    # The real servers depending on the protocol
    http_server_class = {'HTTP/1.0': TestingHTTPSServer,  # type: ignore
                         'HTTP/1.1': TestingThreadingHTTPSServer,  # type: ignore
                         }

    # Provides usable defaults since an https server requires both a
    # private key and a certificate to work.
    def __init__(self, request_handler=http_server.TestingHTTPRequestHandler,
                 protocol_version=None,
                 key_file=ssl_certs.build_path('server_without_pass.key'),
                 cert_file=ssl_certs.build_path('server.crt')):
        http_server.HttpServer.__init__(self, request_handler=request_handler,
                                        protocol_version=protocol_version)
        self.key_file = key_file
        self.cert_file = cert_file
        self.temp_files = []

    def create_server(self):
        return self.server_class(
            (self.host, self.port), self.request_handler_class, self,
            self.key_file, self.cert_file)


class HTTPSServer_urllib(HTTPSServer):
    """Subclass of HTTPSServer that gives https+urllib urls.

    This is for use in testing: connections to this server will always go
    through urllib where possible.
    """

    # urls returned by this server should require the urllib client impl
    _url_protocol = 'https+urllib'



# Copyright (C) 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""InterBranch implementation tests for bzr.

These test the conformance of all the interbranch variations to the
expected API including generally applicable corner cases.
Specific tests for individual formats are in the tests for the formats
itself rather than in tests/per_interbranch/*.py.
"""

from typing import Any, List

from breezy import branchbuilder
from breezy.branch import GenericInterBranch, InterBranch
from breezy.tests import TestCaseWithTransport, multiply_tests


def make_scenarios(test_list):
    """Transform the input test list to a list of scenarios.

    :param formats: A list of tuples:
        (interbranch_class, branch_format_from, branch_format_to).
    """
    result = []
    for interbranch_class, branch_format_from, branch_format_to in test_list:
        id = '{},{},{}'.format(interbranch_class.__name__,
                           branch_format_from.__class__.__name__,
                           branch_format_to.__class__.__name__)
        scenario = (id,
                    {
                        "branch_format_from": branch_format_from,
                        "interbranch_class": interbranch_class,
                        "branch_format_to": branch_format_to,
                        })
        result.append(scenario)
    return result


def default_test_list():
    """Generate the default list of interbranch permutations to test."""
    result = []
    # test the default InterBranch between format 6 and the current
    # default format.
    for optimiser_class in InterBranch.iter_optimisers():
        for format_from_test, format_to_test in \
                optimiser_class._get_branch_formats_to_test():
            result.append((optimiser_class, format_from_test, format_to_test))
    # if there are specific combinations we want to use, we can add them
    # here.
    return result


class TestCaseWithInterBranch(TestCaseWithTransport):

    def make_from_branch(self, relpath):
        return self.make_branch(relpath, format=self.branch_format_from._matchingcontroldir)

    def make_from_branch_and_memory_tree(self, relpath):
        """Create a branch on the default transport and a MemoryTree for it."""
        self.assertEqual(
            self.branch_format_from._matchingcontroldir.get_branch_format(),
            self.branch_format_from)
        return self.make_branch_and_memory_tree(
            relpath, format=self.branch_format_from._matchingcontroldir)

    def make_from_branch_and_tree(self, relpath):
        """Create a branch on the default transport and a working tree for it."""
        self.assertEqual(
            self.branch_format_from._matchingcontroldir.get_branch_format(),
            self.branch_format_from)
        return self.make_branch_and_tree(relpath,
                                         format=self.branch_format_from._matchingcontroldir)

    def make_from_branch_builder(self, relpath):
        self.assertEqual(
            self.branch_format_from._matchingcontroldir.get_branch_format(),
            self.branch_format_from)
        return branchbuilder.BranchBuilder(self.get_transport(relpath),
                                           format=self.branch_format_from._matchingcontroldir)

    def make_to_branch(self, relpath):
        self.assertEqual(
            self.branch_format_to._matchingcontroldir.get_branch_format(),
            self.branch_format_to)
        return self.make_branch(relpath, format=self.branch_format_to._matchingcontroldir)

    def make_to_branch_and_memory_tree(self, relpath):
        """Create a branch on the default transport and a MemoryTree for it."""
        self.assertEqual(
            self.branch_format_to._matchingcontroldir.get_branch_format(),
            self.branch_format_to)
        return self.make_branch_and_memory_tree(
            relpath, format=self.branch_format_to._matchingcontroldir)

    def make_to_branch_and_tree(self, relpath):
        """Create a branch on the default transport and a working tree for it."""
        self.assertEqual(
            self.branch_format_to._matchingcontroldir.get_branch_format(),
            self.branch_format_to)
        return self.make_branch_and_tree(relpath,
                                         format=self.branch_format_to._matchingcontroldir)

    def _sprout(self, origdir, to_url, format):
        if format.supports_workingtrees:
            newbranch = self.make_branch(to_url, format=format)
        else:
            newbranch = self.make_branch(to_url + ".branch", format=format)
        origbranch = origdir.open_branch()
        newbranch.repository.fetch(origbranch.repository)
        origbranch.copy_content_into(newbranch)
        if format.supports_workingtrees:
            wt = newbranch.controldir.create_workingtree()
        else:
            wt = newbranch.create_checkout(to_url, lightweight=True)
        return wt

    def sprout_to(self, origdir, to_url):
        """Sprout a bzrdir, using to_format for the new branch."""
        wt = self._sprout(
            origdir, to_url, self.branch_format_to._matchingcontroldir)
        self.assertEqual(wt.branch._format, self.branch_format_to)
        return wt.controldir

    def sprout_from(self, origdir, to_url):
        """Sprout a bzrdir, using from_format for the new bzrdir."""
        wt = self._sprout(origdir, to_url,
                          self.branch_format_from._matchingcontroldir)
        self.assertEqual(wt.branch._format, self.branch_format_from)
        return wt.controldir


class StubWithFormat:
    """A stub object used to check that convenience methods call Inter's."""

    _format = object()


class StubMatchingInter:
    """An inter for tests.

    This is not a subclass of InterBranch so that missing methods are caught
    and added rather than actually trying to do something.
    """

    _uses: List[Any] = []

    def __init__(self, source, target):
        self.source = source
        self.target = target

    @classmethod
    def is_compatible(klass, source, target):
        return StubWithFormat._format in (source._format, target._format)

    def copy_content_into(self, *args, **kwargs):
        self.__class__._uses.append(
            (self, 'copy_content_into', args, kwargs))


def load_tests(loader, standard_tests, pattern):
    submod_tests = loader.loadTestsFromModuleNames([
        'breezy.tests.per_interbranch.test_fetch',
        'breezy.tests.per_interbranch.test_get',
        'breezy.tests.per_interbranch.test_copy_content_into',
        'breezy.tests.per_interbranch.test_pull',
        'breezy.tests.per_interbranch.test_push',
        ])
    scenarios = make_scenarios(default_test_list())
    return multiply_tests(submod_tests, scenarios, standard_tests)



# Copyright (C) 2008, 2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for commands related to hooks"""

from breezy.branch import Branch
from breezy.tests import TestCaseWithTransport


def _foo_hook():
    pass


class TestHooks(TestCaseWithTransport):

    def _check_hooks_output(self, command_output, hooks):
        for hook_type in Branch.hooks:
            s = "\n  ".join(hooks.get(hook_type, ["<no hooks installed>"]))
            self.assertTrue("{}:\n    {}".format(hook_type, s) in command_output)

    def test_hooks_with_no_hooks(self):
        self.make_branch('.')
        out, err = self.run_bzr('hooks')
        self.assertEqual(err, "")
        for hook_type in Branch.hooks:
            self._check_hooks_output(out, {})

    def test_hooks_with_unnamed_hook(self):
        self.make_branch('.')

        def foo(): return
        Branch.hooks.install_named_hook('post_push', foo, None)
        out, err = self.run_bzr('hooks')
        self._check_hooks_output(out, {'post_push': ["No hook name"]})

    def test_hooks_with_named_hook(self):
        self.make_branch('.')

        def foo(): return
        name = "Foo Bar Hook"
        Branch.hooks.install_named_hook('post_push', foo, name)
        out, err = self.run_bzr('hooks')
        self._check_hooks_output(out, {'post_push': [name]})

    def test_hooks_no_branch(self):
        self.run_bzr('hooks')

    def test_hooks_lazy_with_unnamed_hook(self):
        self.make_branch('.')

        def foo(): return
        Branch.hooks.install_named_hook_lazy('post_push',
                                             'breezy.tests.blackbox.test_hooks',
                                             '_foo_hook',
                                             None)
        out, err = self.run_bzr('hooks')
        self._check_hooks_output(out, {'post_push': ["No hook name"]})

    def test_hooks_lazy_with_named_hook(self):
        self.make_branch('.')

        def foo(): return
        Branch.hooks.install_named_hook_lazy('post_push',
                                             'breezy.tests.blackbox.test_hooks',
                                             '_foo_hook',
                                             'hook has a name')
        out, err = self.run_bzr('hooks')
        self._check_hooks_output(out, {'post_push': ["hook has a name"]})



# Copyright (C) 2005 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .errors import BzrError
from .inventory import Inventory

START_MARK = "# bzr inventory format 3\n"
END_MARK = "# end of inventory\n"


def escape(s):
    """Very simple URL-like escaping.

    (Why not just use backslashes?  Because then we couldn't parse
    lines just by splitting on spaces.)"""
    return (s.replace('\\', r'\x5c')
            .replace(' ', r'\x20')
            .replace('\t', r'\x09')
            .replace('\n', r'\x0a'))


def unescape(s):
    if s.find(' ') != -1:
        raise AssertionError()
    s = (s.replace(r'\x20', ' ')
         .replace(r'\x09', '\t')
         .replace(r'\x0a', '\n')
         .replace(r'\x5c', '\\'))

    # TODO: What if there's anything else?

    return s


def write_text_inventory(inv, outf):
    """Write out inv in a simple trad-unix text format."""
    outf.write(START_MARK)
    for path, ie in inv.iter_entries():
        if inv.is_root(ie.file_id):
            continue

        outf.write(ie.file_id + ' ')
        outf.write(escape(ie.name) + ' ')
        outf.write(ie.kind + ' ')
        outf.write(ie.parent_id + ' ')

        if ie.kind == 'file':
            outf.write(ie.text_id)
            outf.write(' ' + ie.text_sha1)
            outf.write(' ' + str(ie.text_size))
        outf.write("\n")
    outf.write(END_MARK)


def read_text_inventory(tf):
    """Return an inventory read in from tf"""
    if tf.readline() != START_MARK:
        raise BzrError("missing start mark")

    inv = Inventory()

    for l in tf:
        fields = l.split(' ')
        if fields[0] == '#':
            break
        ie = {'file_id': fields[0],
              'name': unescape(fields[1]),
              'kind': fields[2],
              'parent_id': fields[3]}
        # inv.add(ie)

    if l != END_MARK:
        raise BzrError("missing end mark")
    return inv



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the extra cases that WorkingTree.walkdirs can encounter."""

import os

from breezy.tests.features import SymlinkFeature
from breezy.tests.per_workingtree import TestCaseWithWorkingTree

# tests to write:
# type mismatches - file to link, dir, dir to file, link, link to file, dir


class DirBlock:
    """Object representation of the tuples returned by dirstate."""

    def __init__(self, tree, file_path, file_name=None,
                 inventory_kind=None, stat=None, disk_kind='unknown'):
        self.file_path = file_path
        self.abspath = tree.abspath(file_path)
        self.relpath = tree.relpath(file_path)
        if file_name is None:
            file_name = os.path.split(file_path)[-1]
            if len(file_name) == 0:
                file_name = os.path.split(file_path)[-2]
        self.file_name = file_name
        self.inventory_kind = inventory_kind
        self.stat = stat
        self.disk_kind = disk_kind

    def as_tuple(self):
        return (self.relpath, self.file_name, self.disk_kind,
                self.stat, self.inventory_kind)

    def as_dir_tuple(self):
        return self.relpath

    def __str__(self):
        return """
file_path      = {!r}
abspath        = {!r}
relpath        = {!r}
file_name      = {!r}
inventory_kind = {!r}
stat           = {!r}
disk_kind      = {!r}""".format(self.file_path, self.abspath, self.relpath,
                          self.file_name, self.inventory_kind, self.stat,
                          self.disk_kind)


class TestWalkdirs(TestCaseWithWorkingTree):

    added = 'added'
    missing = 'missing'
    unknown = 'unknown'

    def get_tree(self, file_status, prefix=None):
        tree = self.make_branch_and_tree('.')
        dirblocks = []
        paths = [
            file_status + ' file',
            file_status + ' dir/',
            file_status + ' dir/a file',
            file_status + ' empty dir/',
            ]
        self.build_tree(paths)

        def add_dirblock(path, kind):
            dirblock = DirBlock(tree, path)
            if file_status != self.unknown:
                dirblock.inventory_kind = kind
            if file_status != self.missing:
                dirblock.disk_kind = kind
                dirblock.stat = os.lstat(dirblock.relpath)
            dirblocks.append(dirblock)

        add_dirblock(paths[0], 'file')
        add_dirblock(paths[1], 'directory')
        add_dirblock(paths[2], 'file')
        add_dirblock(paths[3], 'directory')

        if file_status != self.unknown:
            tree.add(paths)

        if file_status == self.missing:
            # now make the files be missing
            tree.controldir.root_transport.delete(dirblocks[0].relpath)
            tree.controldir.root_transport.delete_tree(dirblocks[1].relpath)
            tree.controldir.root_transport.delete_tree(dirblocks[3].relpath)

        expected_dirblocks = [
            ('',
             [dirblocks[1].as_tuple()] +
             ([dirblocks[3].as_tuple()]
                 if (tree.has_versioned_directories() or file_status == self.unknown) else []) +
             [dirblocks[0].as_tuple()]
             ),
            (dirblocks[1].as_dir_tuple(),
             [dirblocks[2].as_tuple()]
             ),
            ]
        if (tree.has_versioned_directories() or
                file_status != self.missing):
            expected_dirblocks.append((dirblocks[3].as_dir_tuple(), []))
        if prefix:
            expected_dirblocks = [e for e in expected_dirblocks
                                  if len(e) > 0 and len(e[0]) > 0 and e[0] == prefix]
        return tree, expected_dirblocks

    def _test_walkdir(self, file_status, prefix=""):
        result = []
        tree, expected_dirblocks = self.get_tree(file_status, prefix)
        with tree.lock_read():
            for dirpath, dirblock in tree.walkdirs(prefix):
                result.append((dirpath, list(dirblock)))

        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            result_pos = []
            if len(result) > pos:
                result_pos = result[pos]
            self.assertEqual(item, result_pos)
        self.assertEqual(expected_dirblocks, result)

    def test_walkdir_unknowns(self):
        """unknown files and directories should be reported by walkdirs."""
        self._test_walkdir(self.unknown)

    def test_walkdir_from_unknown_dir(self):
        """Doing a walkdir when the requested prefix is unknown but on disk."""
        self._test_walkdir(self.unknown, 'unknown dir')

    def test_walkdir_missings(self):
        """missing files and directories should be reported by walkdirs."""
        self._test_walkdir(self.missing)

    def test_walkdir_from_dir(self):
        """Doing a walkdir when the requested prefix is known and on disk."""
        self._test_walkdir(self.added, 'added dir')

    def test_walkdir_from_empty_dir(self):
        """Doing a walkdir when the requested prefix is empty dir."""
        self._test_walkdir(self.added, 'added empty dir')

    def test_walkdir_from_missing_dir(self):
        """Doing a walkdir when the requested prefix is missing but on disk."""
        self._test_walkdir(self.missing, 'missing dir')

    def test_walkdirs_type_changes(self):
        """Walkdir shows the actual kinds on disk and the recorded kinds."""
        self.requireFeature(SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree('.')
        paths = ['file1', 'file2', 'dir1/', 'dir2/']
        self.build_tree(paths)
        tree.add(paths)
        tt = tree.transform()
        root_transaction_id = tt.trans_id_tree_path('')
        tt.new_symlink('link1',
                       root_transaction_id, 'link-target', b'link1')
        tt.new_symlink('link2',
                       root_transaction_id, 'link-target', b'link2')
        tt.apply()
        tree.controldir.root_transport.delete_tree('dir1')
        tree.controldir.root_transport.delete_tree('dir2')
        tree.controldir.root_transport.delete('file1')
        tree.controldir.root_transport.delete('file2')
        tree.controldir.root_transport.delete('link1')
        tree.controldir.root_transport.delete('link2')
        changed_paths = ['dir1', 'file1/', 'link1', 'link2/']
        self.build_tree(changed_paths)
        os.symlink('target', 'dir2')
        os.symlink('target', 'file2')
        dir1_stat = os.lstat('dir1')
        dir2_stat = os.lstat('dir2')
        file1_stat = os.lstat('file1')
        file2_stat = os.lstat('file2')
        link1_stat = os.lstat('link1')
        link2_stat = os.lstat('link2')
        expected_dirblocks = [
            ('',
             [('dir1', 'dir1', 'file', dir1_stat,
               'directory' if tree.has_versioned_directories() else None),
              ('dir2', 'dir2', 'symlink', dir2_stat,
               'directory' if tree.has_versioned_directories() else None),
              ('file1', 'file1', 'directory', file1_stat,
               'file'),
              ('file2', 'file2', 'symlink', file2_stat, 'file'),
              ('link1', 'link1', 'file', link1_stat, 'symlink'),
              ('link2', 'link2', 'directory', link2_stat, 'symlink'),
              ]
             )]
        if tree.has_versioned_directories():
            expected_dirblocks.extend([
                ('dir1', []),
                ('dir2', [])])
        expected_dirblocks.extend([
            ('file1', []),
            ('link2', []),
            ])
        with tree.lock_read():
            result = list(tree.walkdirs())
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))

    def test_walkdirs_type_changes_wo_symlinks(self):
        # similar to test_walkdirs_type_changes
        # but don't use symlinks for safe testing on win32
        tree = self.make_branch_and_tree('.')
        paths = ['file1', 'dir1/']
        self.build_tree(paths)
        tree.add(paths)
        tree.controldir.root_transport.delete_tree('dir1')
        tree.controldir.root_transport.delete('file1')
        changed_paths = ['dir1', 'file1/']
        self.build_tree(changed_paths)
        dir1_stat = os.lstat('dir1')
        file1_stat = os.lstat('file1')
        if tree.has_versioned_directories():
            expected_dirblocks = [
                ('',
                 [('dir1', 'dir1', 'file', dir1_stat, 'directory'),
                  ('file1', 'file1', 'directory', file1_stat, 'file'),
                  ]),
                ('dir1', []),
                ('file1', []),
                ]
        else:
            expected_dirblocks = [
                ('',
                 [('dir1', 'dir1', 'file', dir1_stat, None),
                  ('file1', 'file1', 'directory', file1_stat, 'file'),
                  ]
                 ),
                ('file1', []),
                ]
        with tree.lock_read():
            result = list(tree.walkdirs())
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Weave-era working tree objects."""

from io import BytesIO

from ... import conflicts as _mod_conflicts
from ... import errors, lock, osutils
from ... import revision as _mod_revision
from ... import transport as _mod_transport
from ...bzr import conflicts as _mod_bzr_conflicts
from ...bzr import inventory
from ...bzr import transform as bzr_transform
from ...bzr import xml5
from ...bzr.workingtree_3 import PreDirStateWorkingTree
from ...mutabletree import MutableTree
from ...transport.local import LocalTransport
from ...workingtree import WorkingTreeFormat


def get_conflicted_stem(path):
    for suffix in _mod_bzr_conflicts.CONFLICT_SUFFIXES:
        if path.endswith(suffix):
            return path[:-len(suffix)]


class WorkingTreeFormat2(WorkingTreeFormat):
    """The second working tree format.

    This format modified the hash cache from the format 1 hash cache.
    """

    upgrade_recommended = True

    requires_normalized_unicode_filenames = True

    case_sensitive_filename = "Branch-FoRMaT"

    missing_parent_conflicts = False

    supports_versioned_directories = True

    ignore_filename = '.bzrignore'

    def get_format_description(self):
        """See WorkingTreeFormat.get_format_description()."""
        return "Working tree format 2"

    def _stub_initialize_on_transport(self, transport, file_mode):
        """Workaround: create control files for a remote working tree.

        This ensures that it can later be updated and dealt with locally,
        since BzrDirFormat6 and BzrDirFormat5 cannot represent dirs with
        no working tree.  (See bug #43064).
        """
        sio = BytesIO()
        inv = inventory.Inventory()
        xml5.serializer_v5.write_inventory(inv, sio, working=True)
        sio.seek(0)
        transport.put_file('inventory', sio, file_mode)
        transport.put_bytes('pending-merges', b'', file_mode)

    def initialize(self, a_controldir, revision_id=None, from_branch=None,
                   accelerator_tree=None, hardlink=False):
        """See WorkingTreeFormat.initialize()."""
        if not isinstance(a_controldir.transport, LocalTransport):
            raise errors.NotLocalUrl(a_controldir.transport.base)
        if from_branch is not None:
            branch = from_branch
        else:
            branch = a_controldir.open_branch()
        if revision_id is None:
            revision_id = branch.last_revision()
        with branch.lock_write():
            branch.generate_revision_history(revision_id)
        inv = inventory.Inventory()
        wt = WorkingTree2(a_controldir.root_transport.local_abspath('.'),
                          branch,
                          inv,
                          _internal=True,
                          _format=self,
                          _controldir=a_controldir,
                          _control_files=branch.control_files)
        basis_tree = branch.repository.revision_tree(revision_id)
        if basis_tree.path2id('') is not None:
            wt.set_root_id(basis_tree.path2id(''))
        # set the parent list and cache the basis tree.
        if _mod_revision.is_null(revision_id):
            parent_trees = []
        else:
            parent_trees = [(revision_id, basis_tree)]
        wt.set_parent_trees(parent_trees)
        bzr_transform.build_tree(basis_tree, wt)
        for hook in MutableTree.hooks['post_build_tree']:
            hook(wt)
        return wt

    def __init__(self):
        super().__init__()
        from breezy.plugins.weave_fmt.bzrdir import BzrDirFormat6
        self._matchingcontroldir = BzrDirFormat6()

    def open(self, a_controldir, _found=False):
        """Return the WorkingTree object for a_controldir

        _found is a private parameter, do not use it. It is used to indicate
               if format probing has already been done.
        """
        if not _found:
            # we are being called directly and must probe.
            raise NotImplementedError
        if not isinstance(a_controldir.transport, LocalTransport):
            raise errors.NotLocalUrl(a_controldir.transport.base)
        wt = WorkingTree2(a_controldir.root_transport.local_abspath('.'),
                          _internal=True,
                          _format=self,
                          _controldir=a_controldir,
                          _control_files=a_controldir.open_branch().control_files)
        return wt


class WorkingTree2(PreDirStateWorkingTree):
    """This is the Format 2 working tree.

    This was the first weave based working tree.
     - uses os locks for locking.
     - uses the branch last-revision.
    """

    def __init__(self, basedir, *args, **kwargs):
        super().__init__(basedir, *args, **kwargs)
        # WorkingTree2 has more of a constraint that self._inventory must
        # exist. Because this is an older format, we don't mind the overhead
        # caused by the extra computation here.

        # Newer WorkingTree's should only have self._inventory set when they
        # have a read lock.
        if self._inventory is None:
            self.read_working_inventory()

    def _get_check_refs(self):
        """Return the references needed to perform a check of this tree."""
        return [('trees', self.last_revision())]

    def lock_tree_write(self):
        """See WorkingTree.lock_tree_write().

        In Format2 WorkingTrees we have a single lock for the branch and tree
        so lock_tree_write() degrades to lock_write().

        :return: An object with an unlock method which will release the lock
            obtained.
        """
        self.branch.lock_write()
        try:
            token = self._control_files.lock_write()
            return lock.LogicalLockResult(self.unlock, token)
        except:
            self.branch.unlock()
            raise

    def unlock(self):
        # we share control files:
        if self._control_files._lock_count == 3:
            # do non-implementation specific cleanup
            self._cleanup()
            # _inventory_is_modified is always False during a read lock.
            if self._inventory_is_modified:
                self.flush()
            self._write_hashcache_if_dirty()

        # reverse order of locking.
        try:
            return self._control_files.unlock()
        finally:
            self.branch.unlock()

    def _iter_conflicts(self):
        conflicted = set()
        for path, file_class, file_kind, entry in self.list_files():
            stem = get_conflicted_stem(path)
            if stem is None:
                continue
            if stem not in conflicted:
                conflicted.add(stem)
                yield stem

    def conflicts(self):
        with self.lock_read():
            conflicts = _mod_conflicts.ConflictList()
            for conflicted in self._iter_conflicts():
                text = True
                try:
                    if osutils.file_kind(self.abspath(conflicted)) != "file":
                        text = False
                except _mod_transport.NoSuchFile:
                    text = False
                if text is True:
                    for suffix in ('.THIS', '.OTHER'):
                        try:
                            kind = osutils.file_kind(
                                self.abspath(conflicted + suffix))
                            if kind != "file":
                                text = False
                        except _mod_transport.NoSuchFile:
                            text = False
                        if text is False:
                            break
                ctype = {True: 'text conflict',
                         False: 'contents conflict'}[text]
                conflicts.append(_mod_bzr_conflicts.Conflict.factory(
                    ctype, path=conflicted, file_id=self.path2id(conflicted)))
            return conflicts

    def set_conflicts(self, arg):
        raise errors.UnsupportedOperation(self.set_conflicts, self)

    def add_conflicts(self, arg):
        raise errors.UnsupportedOperation(self.add_conflicts, self)



# Copyright (C) 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the branch open with specific URL policy code."""

from typing import List

from .. import urlutils
from ..branch import Branch
from ..bzr import BzrProber
from ..bzr.branch import BranchReferenceFormat
from ..controldir import ControlDir, ControlDirFormat
from ..errors import NotBranchError, RedirectRequested
from ..transport import (Transport, chroot, get_transport, register_transport,
                         register_transport_proto, transport_list_registry,
                         unregister_transport)
from ..url_policy_open import (BadUrl, BranchLoopError, BranchOpener,
                               BranchReferenceForbidden, WhitelistPolicy,
                               _BlacklistPolicy, open_only_scheme)
from . import TestCase, TestCaseWithTransport


class TestBranchOpenerCheckAndFollowBranchReference(TestCase):
    """Unit tests for `BranchOpener.check_and_follow_branch_reference`."""

    def setUp(self):
        super().setUp()
        BranchOpener.install_hook()

    class StubbedBranchOpener(BranchOpener):
        """BranchOpener that provides canned answers.

        We implement the methods we need to to be able to control all the
        inputs to the `follow_reference` method, which is what is
        being tested in this class.
        """

        def __init__(self, references, policy):
            parent_cls = TestBranchOpenerCheckAndFollowBranchReference
            super(parent_cls.StubbedBranchOpener, self).__init__(policy)
            self._reference_values = {}
            for i in range(len(references) - 1):
                self._reference_values[references[i]] = references[i + 1]
            self.follow_reference_calls = []

        def follow_reference(self, url):
            self.follow_reference_calls.append(url)
            return self._reference_values[url]

    def make_branch_opener(self, should_follow_references, references,
                           unsafe_urls=None):
        policy = _BlacklistPolicy(should_follow_references, unsafe_urls)
        opener = self.StubbedBranchOpener(references, policy)
        return opener

    def test_check_initial_url(self):
        # check_and_follow_branch_reference rejects all URLs that are not
        # allowed.
        opener = self.make_branch_opener(None, [], {'a'})
        self.assertRaises(
            BadUrl, opener.check_and_follow_branch_reference, 'a')

    def test_not_reference(self):
        # When branch references are forbidden,
        # check_and_follow_branch_reference does not raise on non-references.
        opener = self.make_branch_opener(False, ['a', None])
        self.assertEqual(
            'a', opener.check_and_follow_branch_reference('a'))
        self.assertEqual(['a'], opener.follow_reference_calls)

    def test_branch_reference_forbidden(self):
        # check_and_follow_branch_reference raises BranchReferenceForbidden if
        # branch references are forbidden and the source URL points to a
        # branch reference.
        opener = self.make_branch_opener(False, ['a', 'b'])
        self.assertRaises(
            BranchReferenceForbidden,
            opener.check_and_follow_branch_reference, 'a')
        self.assertEqual(['a'], opener.follow_reference_calls)

    def test_allowed_reference(self):
        # check_and_follow_branch_reference does not raise if following
        # references is allowed and the source URL points to a branch reference
        # to a permitted location.
        opener = self.make_branch_opener(True, ['a', 'b', None])
        self.assertEqual(
            'b', opener.check_and_follow_branch_reference('a'))
        self.assertEqual(['a', 'b'], opener.follow_reference_calls)

    def test_check_referenced_urls(self):
        # check_and_follow_branch_reference checks if the URL a reference
        # points to is safe.
        opener = self.make_branch_opener(
            True, ['a', 'b', None], unsafe_urls=set('b'))
        self.assertRaises(
            BadUrl, opener.check_and_follow_branch_reference, 'a')
        self.assertEqual(['a'], opener.follow_reference_calls)

    def test_self_referencing_branch(self):
        # check_and_follow_branch_reference raises BranchReferenceLoopError if
        # following references is allowed and the source url points to a
        # self-referencing branch reference.
        opener = self.make_branch_opener(True, ['a', 'a'])
        self.assertRaises(
            BranchLoopError, opener.check_and_follow_branch_reference, 'a')
        self.assertEqual(['a'], opener.follow_reference_calls)

    def test_branch_reference_loop(self):
        # check_and_follow_branch_reference raises BranchReferenceLoopError if
        # following references is allowed and the source url points to a loop
        # of branch references.
        references = ['a', 'b', 'a']
        opener = self.make_branch_opener(True, references)
        self.assertRaises(
            BranchLoopError, opener.check_and_follow_branch_reference, 'a')
        self.assertEqual(['a', 'b'], opener.follow_reference_calls)


class TrackingProber(BzrProber):
    """Subclass of BzrProber which tracks URLs it has been asked to open."""

    seen_urls: List[str] = []

    @classmethod
    def probe_transport(klass, transport):
        klass.seen_urls.append(transport.base)
        return BzrProber.probe_transport(transport)


class TestBranchOpenerStacking(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        BranchOpener.install_hook()

    def make_branch_opener(self, allowed_urls, probers=None):
        policy = WhitelistPolicy(True, allowed_urls, True)
        return BranchOpener(policy, probers)

    def test_probers(self):
        # Only the specified probers should be used
        b = self.make_branch('branch')
        opener = self.make_branch_opener([b.base], probers=[])
        self.assertRaises(NotBranchError, opener.open, b.base)
        opener = self.make_branch_opener([b.base], probers=[BzrProber])
        self.assertEqual(b.base, opener.open(b.base).base)

    def test_default_probers(self):
        # If no probers are specified to the constructor
        # of BranchOpener, then a safe set will be used,
        # rather than all probers registered in bzr.
        self.addCleanup(ControlDirFormat.unregister_prober, TrackingProber)
        ControlDirFormat.register_prober(TrackingProber)
        # Open a location without any branches, so that all probers are
        # tried.
        # First, check that the TrackingProber tracks correctly.
        TrackingProber.seen_urls = []
        opener = self.make_branch_opener(["."], probers=[TrackingProber])
        self.assertRaises(NotBranchError, opener.open, ".")
        self.assertEqual(1, len(TrackingProber.seen_urls))
        TrackingProber.seen_urls = []
        # And make sure it's registered in such a way that ControlDir.open
        # would use it.
        self.assertRaises(NotBranchError, ControlDir.open, ".")
        self.assertEqual(1, len(TrackingProber.seen_urls))

    def test_allowed_url(self):
        # the opener does not raise an exception for branches stacked on
        # branches with allowed URLs.
        stacked_on_branch = self.make_branch('base-branch', format='1.6')
        stacked_branch = self.make_branch('stacked-branch', format='1.6')
        stacked_branch.set_stacked_on_url(stacked_on_branch.base)
        opener = self.make_branch_opener(
            [stacked_branch.base, stacked_on_branch.base])
        # This doesn't raise an exception.
        opener.open(stacked_branch.base)

    def test_nstackable_repository(self):
        # treats branches with UnstackableRepositoryFormats as
        # being not stacked.
        branch = self.make_branch('unstacked', format='knit')
        opener = self.make_branch_opener([branch.base])
        # This doesn't raise an exception.
        opener.open(branch.base)

    def test_allowed_relative_url(self):
        # passes on absolute urls to check_one_url, even if the
        # value of stacked_on_location in the config is set to a relative URL.
        stacked_on_branch = self.make_branch('base-branch', format='1.6')
        stacked_branch = self.make_branch('stacked-branch', format='1.6')
        stacked_branch.set_stacked_on_url('../base-branch')
        opener = self.make_branch_opener(
            [stacked_branch.base, stacked_on_branch.base])
        # Note that stacked_on_branch.base is not '../base-branch', it's an
        # absolute URL.
        self.assertNotEqual('../base-branch', stacked_on_branch.base)
        # This doesn't raise an exception.
        opener.open(stacked_branch.base)

    def test_allowed_relative_nested(self):
        # Relative URLs are resolved relative to the stacked branch.
        self.get_transport().mkdir('subdir')
        a = self.make_branch('subdir/a', format='1.6')
        b = self.make_branch('b', format='1.6')
        b.set_stacked_on_url('../subdir/a')
        c = self.make_branch('subdir/c', format='1.6')
        c.set_stacked_on_url('../../b')
        opener = self.make_branch_opener([c.base, b.base, a.base])
        # This doesn't raise an exception.
        opener.open(c.base)

    def test_forbidden_url(self):
        # raises a BadUrl exception if a branch is stacked on a
        # branch with a forbidden URL.
        stacked_on_branch = self.make_branch('base-branch', format='1.6')
        stacked_branch = self.make_branch('stacked-branch', format='1.6')
        stacked_branch.set_stacked_on_url(stacked_on_branch.base)
        opener = self.make_branch_opener([stacked_branch.base])
        self.assertRaises(BadUrl, opener.open, stacked_branch.base)

    def test_forbidden_url_nested(self):
        # raises a BadUrl exception if a branch is stacked on a
        # branch that is in turn stacked on a branch with a forbidden URL.
        a = self.make_branch('a', format='1.6')
        b = self.make_branch('b', format='1.6')
        b.set_stacked_on_url(a.base)
        c = self.make_branch('c', format='1.6')
        c.set_stacked_on_url(b.base)
        opener = self.make_branch_opener([c.base, b.base])
        self.assertRaises(BadUrl, opener.open, c.base)

    def test_self_stacked_branch(self):
        # raises StackingLoopError if a branch is stacked on
        # itself. This avoids infinite recursion errors.
        a = self.make_branch('a', format='1.6')
        # Bazaar 1.17 and up make it harder to create branches like this.
        # It's still worth testing that we don't blow up in the face of them,
        # so we grovel around a bit to create one anyway.
        a.get_config().set_user_option('stacked_on_location', a.base)
        opener = self.make_branch_opener([a.base])
        self.assertRaises(BranchLoopError, opener.open, a.base)

    def test_loop_stacked_branch(self):
        # raises StackingLoopError if a branch is stacked in such
        # a way so that it is ultimately stacked on itself. e.g. a stacked on
        # b stacked on a.
        a = self.make_branch('a', format='1.6')
        b = self.make_branch('b', format='1.6')
        a.set_stacked_on_url(b.base)
        b.set_stacked_on_url(a.base)
        opener = self.make_branch_opener([a.base, b.base])
        self.assertRaises(BranchLoopError, opener.open, a.base)
        self.assertRaises(BranchLoopError, opener.open, b.base)

    def test_custom_opener(self):
        # A custom function for opening a control dir can be specified.
        a = self.make_branch('a', format='2a')
        b = self.make_branch('b', format='2a')
        b.set_stacked_on_url(a.base)

        TrackingProber.seen_urls = []
        opener = self.make_branch_opener(
            [a.base, b.base], probers=[TrackingProber])
        opener.open(b.base)
        self.assertEqual(
            set(TrackingProber.seen_urls), {b.base, a.base})

    def test_custom_opener_with_branch_reference(self):
        # A custom function for opening a control dir can be specified.
        a = self.make_branch('a', format='2a')
        b_dir = self.make_controldir('b')
        b = BranchReferenceFormat().initialize(b_dir, target_branch=a)
        TrackingProber.seen_urls = []
        opener = self.make_branch_opener(
            [a.base, b.base], probers=[TrackingProber])
        opener.open(b.base)
        self.assertEqual(
            set(TrackingProber.seen_urls), {b.base, a.base})


class TestRedirects(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        BranchOpener.install_hook()

    def setup_redirect(self, target_url):
        class RedirectingTransport(Transport):

            def get(self, name):
                raise RedirectRequested(self.base, target_url)

            def _redirected_to(self, source, target):
                return get_transport(target)

        register_transport_proto(
            'redirecting://', help="Test transport that redirects.")
        register_transport('redirecting://', RedirectingTransport)
        self.addCleanup(unregister_transport, 'redirecting://', RedirectingTransport)

    def make_branch_opener(self, allowed_urls, probers=None):
        policy = WhitelistPolicy(True, allowed_urls, True)
        return BranchOpener(policy, probers)

    def test_redirect_forbidden(self):
        b = self.make_branch('b')
        self.setup_redirect(b.base)
        class TrackingProber(BzrProber):
            seen_urls = []

            @classmethod
            def probe_transport(klass, transport):
                klass.seen_urls.append(transport.base)
                return BzrProber.probe_transport(transport)

        opener = self.make_branch_opener(['redirecting:///'], probers=[TrackingProber])
        self.assertRaises(BadUrl, opener.open, 'redirecting:///')

        opener = self.make_branch_opener(['redirecting:///', b.base], probers=[TrackingProber])
        opener.open('redirecting:///')


class TestOpenOnlyScheme(TestCaseWithTransport):
    """Tests for `open_only_scheme`."""

    def setUp(self):
        super().setUp()
        BranchOpener.install_hook()

    def test_hook_does_not_interfere(self):
        # The transform_fallback_location hook does not interfere with regular
        # stacked branch access outside of open_only_scheme.
        self.make_branch('stacked')
        self.make_branch('stacked-on')
        Branch.open('stacked').set_stacked_on_url('../stacked-on')
        Branch.open('stacked')

    def get_chrooted_scheme(self, relpath):
        """Create a server that is chrooted to `relpath`.

        :return: ``(scheme, get_url)`` where ``scheme`` is the scheme of the
            chroot server and ``get_url`` returns URLs on said server.
        """
        transport = self.get_transport(relpath)
        chroot_server = chroot.ChrootServer(transport)
        chroot_server.start_server()
        self.addCleanup(chroot_server.stop_server)

        def get_url(relpath):
            return chroot_server.get_url() + relpath

        return (urlutils.URL.from_string(chroot_server.get_url()).scheme,
                get_url)

    def test_stacked_within_scheme(self):
        # A branch that is stacked on a URL of the same scheme is safe to
        # open.
        self.get_transport().mkdir('inside')
        self.make_branch('inside/stacked')
        self.make_branch('inside/stacked-on')
        scheme, get_chrooted_url = self.get_chrooted_scheme('inside')
        Branch.open(get_chrooted_url('stacked')).set_stacked_on_url(
            get_chrooted_url('stacked-on'))
        open_only_scheme(scheme, get_chrooted_url('stacked'))

    def test_stacked_outside_scheme(self):
        # A branch that is stacked on a URL that is not of the same scheme is
        # not safe to open.
        self.get_transport().mkdir('inside')
        self.get_transport().mkdir('outside')
        self.make_branch('inside/stacked')
        self.make_branch('outside/stacked-on')
        scheme, get_chrooted_url = self.get_chrooted_scheme('inside')
        Branch.open(get_chrooted_url('stacked')).set_stacked_on_url(
            self.get_url('outside/stacked-on'))
        self.assertRaises(
            BadUrl, open_only_scheme, scheme, get_chrooted_url('stacked'))
